# Copyright (c) 2022-present, FriendliAI Inc. All rights reserved.

"""PeriFlow Tokenization API."""

# pylint: disable=no-name-in-module

from __future__ import annotations

from typing import List

import requests
from pydantic import ValidationError
from requests import HTTPError

from periflow.errors import APIError, SessionClosedError
from periflow.schema.api.v1.codegen.tokenize_pb2 import V1TokenizeRequest
from periflow.schema.api.v1.tokenize import V1TokenizeResponse
from periflow.sdk.api.base import BaseAPI
from periflow.utils.request import DEFAULT_REQ_TIMEOUT


class Tokenization(BaseAPI):
    """PeriFlow tokenization API."""

    @property
    def _api_path(self) -> str:
        return "v1/tokenize"

    def create(self, prompt: str) -> List[int]:
        """Create a tokenization result.

        Args:
            prompt (str): Input prompt in text.

        Raises:
            APIError: Raised when the input has invalid format.

        Returns:
            List[int]: A list of token IDs.

        """
        request_pb = V1TokenizeRequest(prompt=prompt)
        request_data = request_pb.SerializeToString()
        try:
            response = requests.post(
                url=self._endpoint,
                data=request_data,
                headers=self._get_headers(),
                timeout=DEFAULT_REQ_TIMEOUT,
            )
        except HTTPError as exc:
            raise APIError(str(exc)) from exc

        resp_data = response.json()
        try:
            tokenize_res = V1TokenizeResponse.model_validate(resp_data)
        except ValidationError as exc:
            raise APIError(resp_data) from exc

        return tokenize_res.tokens

    async def acreate(self, prompt: str) -> List[int]:
        """Create a tokenization result asynchronously.

        :::info
        You must open API session with `api_session()` before `acreate()`.
        :::

        Args:
            prompt (str): Input prompt in text.

        Raises:
            APIError: Raised when the input has invalid format.

        Returns:
            List[int]: A list of token IDs.

        """
        if self._session is None:
            raise SessionClosedError("Create a session with 'api_session' first.")

        request_pb = V1TokenizeRequest(prompt=prompt)
        request_data = request_pb.SerializeToString()
        response = await self._session.post(url=self._endpoint, data=request_data)

        if 400 <= response.status < 500:
            raise APIError(
                f"{response.status} Client Error: {response.reason} for url: {self._endpoint}"
            )
        if 500 <= response.status < 600:
            raise APIError(
                f"{response.status} Server Error: {response.reason} for url: {self._endpoint}"
            )

        resp_data = await response.json()
        try:
            tokenize_res = V1TokenizeResponse.model_validate(resp_data)
        except ValidationError as exc:
            raise APIError(resp_data) from exc

        return tokenize_res.tokens



# Copyright (c) 2022-present, FriendliAI Inc. All rights reserved.

"""PeriFlow Deployment Configurator."""

from __future__ import annotations

import json
from typing import Any, Dict

from periflow.configurator.base import IO, Configurator
from periflow.utils.format import secho_error_and_exit


class DRCConfigurator(Configurator):
    """Configurator for default request config."""

    @property
    def validation_schema(self) -> dict:
        """JSON schema to validate DRC file contents."""
        return {
            "type": "object",
            "properties": {
                "stop": {"type": "array", "items": {"type": "string"}},
                "stop_tokens": {
                    "type": "object",
                    "properties": {
                        "properties": {
                            "tokens": {"type": "array", "items": {"type": "integer"}}
                        },
                        "required": ["tokens"],
                    },
                },
                "bad_words": {"type": "array", "items": {"type": "string"}},
                "bad_word_tokens": {
                    "type": "object",
                    "properties": {
                        "properties": {
                            "tokens": {"type": "array", "items": {"type": "integer"}}
                        },
                        "required": ["tokens"],
                    },
                },
            },
            "allOf": [
                {"not": {"required": ["stop", "stop_tokens"]}},
                {"not": {"required": ["bad_words", "bad_word_tokens"]}},
            ],
            "minProperties": 1,
            "additionalProperties": False,
        }

    @classmethod
    def from_file(cls, f: IO) -> Configurator:
        """Build a DRC configurator from a DRC file."""
        try:
            config: Dict[str, Any] = json.load(f)
        except json.JSONDecodeError as e:
            secho_error_and_exit(f"Error occurred while parsing config file: {e!r}")

        return cls(config)  # type: ignore


class DeploymentConfigurator(Configurator):
    """Deployment configurator."""

    @classmethod
    def from_file(cls, f: IO) -> Configurator:
        """Build a DRC configurator from a DRC file."""
        try:
            config: Dict[str, Any] = json.load(f)
        except json.JSONDecodeError as e:
            secho_error_and_exit(f"Error occurred while parsing config file: {e!r}")

        return cls(config)  # type: ignore


class OrcaDeploymentConfigurator(DeploymentConfigurator):
    """Orca deployment config."""

    @property
    def validation_schema(self) -> dict:
        """JSON schema to validate config."""
        return {
            "type": "object",
            "properties": {
                "max_batch_size": {"type": "integer"},
                "max_token_count": {"type": "integer"},
                "max_num_tokens_to_replace": {"type": "integer"},
            },
            "minProperties": 1,
            "additionalProperties": False,
        }



# Copyright (c) 2022-present, FriendliAI Inc. All rights reserved.

"""PeriFlow Quantizer Maps."""

from __future__ import annotations

from typing import Any, Dict, Type

from periflow.enums import QuantMode
from periflow.errors import NotSupportedQuantModeError
from periflow.modules.converter.base import OneOfConverter
from periflow.modules.quantizer.awq.base import AWQHook, AWQQuantizer
from periflow.modules.quantizer.awq.models.llama import AWQLlamaHook
from periflow.modules.quantizer.awq.models.mpt import AWQMPTHook
from periflow.modules.quantizer.base import CommonQuantizer
from periflow.modules.quantizer.schema.config import OneOfQuantConfig
from periflow.modules.quantizer.smoothquant.base import (
    SmoothQuantHook,
    SmoothQuantQuantizer,
)
from periflow.modules.quantizer.smoothquant.models.bloom import SmoothQuantBloomHook
from periflow.modules.quantizer.smoothquant.models.codegen import SmoothQuantCodeGenHook
from periflow.modules.quantizer.smoothquant.models.falcon import SmoothQuantFalconHook
from periflow.modules.quantizer.smoothquant.models.gpt2 import SmoothQuantGPT2Hook
from periflow.modules.quantizer.smoothquant.models.gpt_neox import (
    SmoothQuantGPTNeoXHook,
)
from periflow.modules.quantizer.smoothquant.models.gptj import SmoothQuantGPTJHook
from periflow.modules.quantizer.smoothquant.models.llama import SmoothQuantLlamaHook
from periflow.modules.quantizer.smoothquant.models.mpt import SmoothQuantMPTHook
from periflow.modules.quantizer.smoothquant.models.opt import SmoothQuantOPTHook

model_arch_smoothquant_hook_map: Dict[str, type[SmoothQuantHook]] = {
    "OPTForCausalLM": SmoothQuantOPTHook,
    "MPTForCausalLM": SmoothQuantMPTHook,
    "BloomForCausalLM": SmoothQuantBloomHook,
    "CodeGenForCausalLM": SmoothQuantCodeGenHook,
    "GPTNeoXForCausalLM": SmoothQuantGPTNeoXHook,
    "GPTJForCausalLM": SmoothQuantGPTJHook,
    "GPT2LMHeadModel": SmoothQuantGPT2Hook,
    "FalconForCausalLM": SmoothQuantFalconHook,
    "LlamaForCausalLM": SmoothQuantLlamaHook,
}

model_arch_awq_hook_map: Dict[str, type[AWQHook]] = {
    "MPTForCausalLM": AWQMPTHook,
    "LlamaForCausalLM": AWQLlamaHook,
}


def get_quanthook_map(quant_mode: QuantMode) -> Dict[str, Any]:
    """Get quantizer map."""
    if quant_mode == QuantMode.SMOOTH_QUANT:
        return model_arch_smoothquant_hook_map
    if quant_mode == QuantMode.AWQ:
        return model_arch_awq_hook_map
    raise NotSupportedQuantModeError(
        invalid_option=quant_mode,
        valid_options=[e.value for e in QuantMode],
    )


def get_quantizer_class(quant_mode: QuantMode) -> Type[CommonQuantizer]:
    """Get quantizer class."""
    if quant_mode == QuantMode.SMOOTH_QUANT:
        return SmoothQuantQuantizer
    if quant_mode == QuantMode.AWQ:
        return AWQQuantizer
    raise NotSupportedQuantModeError(
        invalid_option=quant_mode,
        valid_options=[e.value for e in QuantMode],
    )


def get_quantized_converter(
    model_arch: str,
    quant_config: OneOfQuantConfig,
    converter: OneOfConverter,
) -> CommonQuantizer:
    """Get quantizer for specific model architecture with quant mode and args."""
    quant_mode = quant_config.mode
    quantizer = get_quantizer_class(quant_mode)
    quanthook_map = get_quanthook_map(quant_mode)
    quanthook = quanthook_map[model_arch](quant_config, converter)
    return quantizer(quanthook, quant_config, converter)



# Copyright (c) 2022-present, FriendliAI Inc. All rights reserved.

"""PeriFlow AWQ Quantizer Base."""

from __future__ import annotations

import gc
from abc import abstractmethod
from dataclasses import fields
from typing import Any, Callable, Dict, Iterator, List, Tuple, Type, cast

import datasets  # type: ignore[import]
import numpy as np
import torch
from datasets.utils.logging import disable_progress_bar  # type: ignore[import]
from tqdm import tqdm

from periflow.errors import QuantizationError
from periflow.modules.converter.utils import get_tokenizer, nontype_partial
from periflow.modules.quantizer.awq.utils import (
    apply_module_clip,
    apply_module_scale,
    search_module_clip,
    search_module_scale,
)
from periflow.modules.quantizer.base import AbstractQuantHook, CommonQuantizer
from periflow.modules.quantizer.schema.config import AWQConfig
from periflow.modules.quantizer.schema.data import (
    ModuleName,
    QuantInput,
    TFQuantInputs,
    TFQuantResults,
    WeightOnlyQuantResult,
)
from periflow.modules.quantizer.utils import (
    collect_inps,
    get_weight_only_quant_scales,
    quantized_linear_weight_convert,
    quantized_qkv_weight_convert,
    safe_load_datasets,
    scale_convert,
)


class AWQScaler(torch.nn.Module):
    """Store AWQ scale before linear layers.

    If the linear layer is quantized, but the previous layer can't be scaled,
    then we need to store the AWQ scale in a separate module. This module
    is used to store the AWQ scale.
    """

    def __init__(self, in_dim: int):
        """Initialize AWQScaler."""
        super().__init__()
        self.scale = torch.nn.Parameter(torch.ones(in_dim))

    def forward(self, x):
        """Scale input by AWQ scale."""
        return (x / self.scale.view(1, 1, -1)).to(x.dtype)


class AWQHook(AbstractQuantHook):
    """Quantization Hook for SmoothQuant."""

    @abstractmethod
    def iter_inspect_modules(
        self, model: torch.nn.Module, model_kwargs: Dict[str, Any]
    ) -> Iterator[
        Tuple[
            List[torch.nn.Module],
            List[torch.nn.Module],
            ModuleName,
            torch.nn.Module,
            Dict[str, Any],
        ]
    ]:
        """Returns iterator of modules to inspect for AWQ scale."""

    @abstractmethod
    def add_pre_scaler(
        self,
        model: torch.nn.Module,
    ) -> torch.nn.Module:
        """Add scaler for storing AWQ scale in modules."""

    @abstractmethod
    def get_inspect_module_types(
        self, block: torch.nn.Module
    ) -> Tuple[Type[torch.nn.Module], ...]:
        """Returns the type of inspect modules in transformer block."""

    def _register_pre_scaler(
        self,
        linear: torch.nn.Module,
    ) -> AWQScaler:
        """Register pre-scaler for storing AWQ scale in modules."""
        scaler = AWQScaler(linear.in_features)  # type: ignore

        def pre_scaler_hook(_, x: Tuple[Any, ...]) -> Tuple[torch.Tensor, ...]:
            return (scaler(x[0]),)

        linear.register_forward_pre_hook(pre_scaler_hook)
        return scaler

    def get_quant_result(
        self,
        quant_input: TFQuantInputs,
        **kwargs: Any,
    ) -> TFQuantResults:
        """Get quantization result for AWQ."""
        awq_args = cast(AWQConfig, self.quant_config).awq_args

        def get_scale(
            quant_input: QuantInput,
        ) -> WeightOnlyQuantResult:
            weight, name, start, end = (
                quant_input.weight,
                quant_input.name,
                quant_input.start_offset,
                quant_input.end_offset,
            )

            return get_weight_only_quant_scales(
                layer_name=name,
                w=weight[start:end],
                q_bit=awq_args.quant_bit,
                q_group_size=awq_args.quant_group_size,
            )

        return TFQuantResults(
            layer_prefix_with_index=f"{self.quantized_layer_prefix}.{quant_input.layer_index}",
            q=get_scale(quant_input.q),
            k=get_scale(quant_input.k),
            v=get_scale(quant_input.v),
            attn_fc=get_scale(quant_input.attn_fc),
            ff1=get_scale(quant_input.ff1),
            ff2=get_scale(quant_input.ff2),
        )

    @property
    @abstractmethod
    def avoid_clipping_layer_names(self) -> List[str]:
        """Return the layer names to avoid clipping."""

    @property
    def modified_layers_convert_dict(
        self,
    ) -> Dict[str, Callable[[Dict[str, torch.Tensor], str], np.ndarray]]:
        """Return the convert_dict for modified layers."""
        return {
            "attn/c_proj/awq/pre_scale:0": nontype_partial(
                scale_convert,
                per_layer_postfixes=[".attn.scaler.scale"],
                data_type="fp32",
            ),
            "mlp/c_proj/awq/pre_scale:0": nontype_partial(
                scale_convert,
                per_layer_postfixes=[".ffn.scaler.scale"],
                data_type="fp32",
            ),
        }

    @property
    def quantized_convert_dict(
        self,
    ) -> Dict[str, Callable[[Dict[str, torch.Tensor], str], np.ndarray]]:
        """Return the convert_dict for quantized layers."""
        quant_config = cast(AWQConfig, self.quant_config)
        n_bit = quant_config.awq_args.quant_bit
        return {
            "attn/c_attn/awq/scale:0": nontype_partial(
                scale_convert,
                per_layer_postfixes=[
                    ".q.woq_weight_scale",
                    ".k.woq_weight_scale",
                    ".v.woq_weight_scale",
                ],
                data_type=self.converter.data_type,
            ),
            "attn/c_attn/awq/zero:0": nontype_partial(
                scale_convert,
                per_layer_postfixes=[
                    ".q.woq_weight_zp",
                    ".k.woq_weight_zp",
                    ".v.woq_weight_zp",
                ],
                data_type=self.converter.data_type,
            ),
            "attn/c_attn/awq/weight:0": nontype_partial(
                quantized_qkv_weight_convert,
                per_layer_postfixes=[
                    ".q.woq_weight",
                    ".k.woq_weight",
                    ".v.woq_weight",
                ],
                n_bit=n_bit,
            ),
            "attn/c_proj/awq/scale:0": nontype_partial(
                scale_convert,
                per_layer_postfixes=[".attn_fc.woq_weight_scale"],
                data_type=self.converter.data_type,
            ),
            "attn/c_proj/awq/zero:0": nontype_partial(
                scale_convert,
                per_layer_postfixes=[".attn_fc.woq_weight_zp"],
                data_type=self.converter.data_type,
            ),
            "attn/c_proj/awq/weight:0": nontype_partial(
                quantized_linear_weight_convert,
                per_layer_postfixes=[".attn_fc.woq_weight"],
                n_bit=n_bit,
            ),
            "mlp/c_fc/awq/scale:0": nontype_partial(
                scale_convert,
                per_layer_postfixes=[".ff1.woq_weight_scale"],
                data_type=self.converter.data_type,
            ),
            "mlp/c_fc/awq/zero:0": nontype_partial(
                scale_convert,
                per_layer_postfixes=[".ff1.woq_weight_zp"],
                data_type=self.converter.data_type,
            ),
            "mlp/c_fc/awq/weight:0": nontype_partial(
                quantized_linear_weight_convert,
                per_layer_postfixes=[".ff1.woq_weight"],
                n_bit=n_bit,
            ),
            "mlp/c_proj/awq/scale:0": nontype_partial(
                scale_convert,
                per_layer_postfixes=[".ff2.woq_weight_scale"],
                data_type=self.converter.data_type,
            ),
            "mlp/c_proj/awq/zero:0": nontype_partial(
                scale_convert,
                per_layer_postfixes=[".ff2.woq_weight_zp"],
                data_type=self.converter.data_type,
            ),
            "mlp/c_proj/awq/weight:0": nontype_partial(
                quantized_linear_weight_convert,
                per_layer_postfixes=[".ff2.woq_weight"],
                n_bit=n_bit,
            ),
        }


class AWQQuantizer(CommonQuantizer):
    """Quantizer for AWQ."""

    def check_config(self) -> None:
        """Check if the AWQ quantization config is valid."""
        super().check_config()
        quant_config = cast(AWQConfig, self.quant_config)
        awq_args = quant_config.awq_args
        if awq_args.quant_bit not in [4, 8]:
            raise QuantizationError(
                f"Invalid quant_bit {awq_args.quant_bit} for AWQ."
                "You can only use 4 or 8 bit for AWQ."
            )
        if awq_args.quant_group_size not in [64]:
            raise QuantizationError(
                f"Invalid quant_group_size {awq_args.quant_group_size} for AWQ."
                "You can only use 64 for AWQ."
            )

    def get_calib_dataset(self) -> datasets.Dataset:
        """Get calibration dataset for AWQ."""
        data_cfg = self.quant_config.calibration_dataset
        tokenizer = get_tokenizer(self.converter.config.name_or_path)
        dataset = safe_load_datasets(data_cfg)

        def preprocess(sample) -> Dict[str, Any]:
            """Preprocess dataset for AWQ."""
            return {"input_ids": tokenizer(sample).input_ids}

        disable_progress_bar()
        dataset = (
            dataset.shuffle(self.quant_config.seed)
            .filter(
                lambda sample: len(tokenizer(sample)) <= data_cfg.max_length,
                input_columns=data_cfg.lookup_column_name,
            )
            .map(function=preprocess, input_columns=data_cfg.lookup_column_name)
            .filter(
                lambda sample: torch.tensor(sample).numel() != 0,
                input_columns="input_ids",
            )
            .select(range(data_cfg.num_samples))
        )

        return dataset

    def get_batched_samples(self):
        """Get batched samples from dataset."""
        dataset = self.get_calib_dataset()

        samples = []
        for sample in dataset["input_ids"]:
            samples.append(torch.tensor(sample))

        seqlen = self.quant_config.calibration_dataset.max_length
        batched_samples = torch.cat(samples)
        if len(batched_samples) // seqlen == 0:
            return batched_samples.unsqueeze(0)

        batched_samples = [
            batched_samples[i * seqlen : (i + 1) * seqlen].unsqueeze(0)
            for i in range(len(batched_samples) // seqlen)
        ]
        batched_samples = torch.cat(batched_samples, dim=0)

        return batched_samples

    def _apply_awq_scale_clip_block(
        self,
        block: torch.nn.Module,
        block_inps: torch.Tensor,
        block_kwargs: Dict[str, Any],
    ) -> None:
        """Search AWQ scale, clipping range and Apply them into a transformer block."""
        # pylint: disable=too-many-locals

        inpsected_mod_types = cast(AWQHook, self.hook).get_inspect_module_types(block)
        input_feat, intra_kwargs = collect_inps(
            block_inps,
            tuple([*self.hook.get_linear_layer_types(), *inpsected_mod_types]),
            block,
            block_kwargs,
        )
        awq_args = cast(AWQConfig, self.quant_config).awq_args

        for prev_ops, linear_layers, layer_name, module2inspect, kwargs in cast(
            AWQHook, self.hook
        ).iter_inspect_modules(block, intra_kwargs):
            module_inp = input_feat[layer_name]
            scales = search_module_scale(
                module2inspect,
                linear_layers,
                module_inp,
                awq_args.quant_group_size,
                awq_args.quant_bit,
                module_kwargs=kwargs,
            )
            scaled_module_inp = apply_module_scale(
                prev_ops,
                linear_layers,
                scales.to(self.quant_config.device),
                module_inp.to(self.quant_config.device),
            )

            if any(
                (
                    avoid in layer_name
                    for avoid in cast(AWQHook, self.hook).avoid_clipping_layer_names
                )
            ):
                continue
            max_val = search_module_clip(
                linear_layers,
                scaled_module_inp,
                awq_args.quant_group_size,
                awq_args.quant_bit,
                n_sample_token=self.quant_config.calibration_dataset.num_samples,
            )
            apply_module_clip(
                max_val.to(self.quant_config.device),
                linear_layers,
            )

        del input_feat

    @torch.no_grad()
    def _apply_awq_scale_clip(
        self,
        model: torch.nn.Module,
    ) -> None:
        """Search AWQ scale, clipping range and Apply them into model."""
        # pylint: disable=too-many-locals
        model.eval()
        model.to(self.quant_config.device)
        tf_blocks = self.hook.get_tf_blocks(model)
        block_inps, block_kwargs = collect_inps(
            inps=self.get_batched_samples(),
            target_classes=(type(tf_blocks[0]),),
            module=model,
            module_kwargs={},
        )
        model.to("cpu")
        gc.collect()
        torch.cuda.empty_cache()

        for block, inps, kwargs in tqdm(
            zip(
                tf_blocks,
                block_inps.values(),
                block_kwargs.values(),
            ),
            total=len(tf_blocks),
            desc="Search and Apply AWQ Scale, Clip range..",
        ):
            block = block.to(self.quant_config.device)
            self._apply_awq_scale_clip_block(block, inps, kwargs)
            block.to("cpu")
            gc.collect()
            torch.cuda.empty_cache()

    @torch.no_grad()
    def pre_quantize(
        self,
        model: torch.nn.Module,
    ) -> None:
        """Pre-procedure that should be called before quantize() is called."""
        model = cast(AWQHook, self.hook).add_pre_scaler(model)
        self._apply_awq_scale_clip(model)

    @torch.no_grad()
    def quantize(
        self,
        model: torch.nn.Module,
    ) -> Iterator[TFQuantResults]:
        """Quantize model with SmoothQuant."""
        model.eval()
        model.to(self.quant_config.device)
        for quant_input in tqdm(
            self.hook.iter_quant_inputs(model),
            total=len(self.hook.get_tf_blocks(model)),
            desc="Quantize model..",
        ):
            yield cast(AWQHook, self.hook).get_quant_result(
                quant_input, quant_config=cast(AWQConfig, self.quant_config)
            )

    def get_quantized_state_dict(
        self, model: torch.nn.Module, quant_result_iter: Iterator[TFQuantResults]
    ) -> Dict[str, torch.Tensor]:
        """Get quantized state dict for AWQ."""
        state_dict = model.state_dict()
        for quant_result in quant_result_iter:
            for field in fields(quant_result):
                layer_name = field.name
                layer = getattr(quant_result, layer_name)

                if isinstance(layer, WeightOnlyQuantResult):
                    state_dict[
                        f"{quant_result.layer_prefix_with_index}.{layer_name}.woq_weight"
                    ] = layer.q_weight
                    state_dict[
                        f"{quant_result.layer_prefix_with_index}.{layer_name}.woq_weight_scale"
                    ] = layer.weight_scale
                    state_dict[
                        f"{quant_result.layer_prefix_with_index}.{layer_name}.woq_weight_zp"
                    ] = layer.zero_point

        return state_dict



# Copyright (c) 2022-present, FriendliAI Inc. All rights reserved.

"""PeriFlow V1 Tokenize API Schemas."""

from __future__ import annotations

from typing import List, Optional

from pydantic import BaseModel


class V1TokenizeRequest(BaseModel):
    """V1 tokenize request schema."""

    prompt: Optional[str] = None


class V1TokenizeResponse(BaseModel):
    """V1 tokenize response schema."""

    tokens: List[int]



# Copyright (c) 2022-present, FriendliAI Inc. All rights reserved.

"""PeriFlow User Clients."""

from __future__ import annotations

from string import Template
from typing import Any, Dict, List
from uuid import UUID

from periflow.client.base import Client, GroupRequestMixin, UserRequestMixin
from periflow.enums import GroupRole, ProjectRole


class UserMFAClient(Client):
    """User MFA client."""

    @property
    def url_path(self) -> Template:
        """Get an URL path."""
        return Template(self.url_provider.get_auth_uri("mfa"))

    def initiate_mfa(self, mfa_type: str, mfa_token: str) -> None:
        """Authenticate by MFA token."""
        self.bare_post(path=f"challenge/{mfa_type}", headers={"x-mfa-token": mfa_token})


class UserSignUpClient(Client):
    """User sign-up client."""

    @property
    def url_path(self) -> Template:
        """Get an URL path."""
        return Template(self.url_provider.get_auth_uri("pf_user/self_signup"))

    def verify(self, token: str, key: str) -> None:
        """Verify the email account with the token to sign up."""
        self.bare_post(path="confirm", json={"email_token": token, "key": key})


class UserClient(Client, UserRequestMixin):
    """User client."""

    def __init__(self, **kwargs) -> None:
        """Initialize user client."""
        self.initialize_user()
        super().__init__(**kwargs)

    @property
    def url_path(self) -> Template:
        """Get an URL path."""
        return Template(self.url_provider.get_auth_uri("pf_user"))

    def change_password(self, old_password: str, new_password: str) -> None:
        """Change password."""
        self.update(
            pk=self.user_id,
            path="password",
            json={"old_password": old_password, "new_password": new_password},
        )

    def set_group_privilege(
        self, pf_group_id: UUID, pf_user_id: UUID, privilege_level: GroupRole
    ) -> None:
        """Update user role in the orgnaization."""
        self.partial_update(
            pk=pf_user_id,
            path=f"pf_group/{pf_group_id}/privilege_level",
            json={"privilege_level": privilege_level.value},
        )

    def get_project_membership(self, pf_project_id: UUID) -> Dict[str, Any]:
        """Get the project membership info of the user."""
        data = self.retrieve(
            pk=self.user_id,
            path=f"pf_project/{pf_project_id}",
        )
        return data

    def add_to_project(
        self, pf_user_id: UUID, pf_project_id: UUID, access_level: ProjectRole
    ) -> None:
        """Add a new member to a project."""
        self.post(
            path=f"{pf_user_id}/pf_project/{pf_project_id}",
            json={"access_level": access_level.value},
        )

    def delete_from_org(self, pf_user_id: UUID, pf_org_id: UUID) -> None:
        """Delete a member from the organization."""
        self.delete(pk=pf_user_id, path=f"pf_group/{pf_org_id}")

    def delete_from_project(self, pf_user_id: UUID, pf_project_id: UUID) -> None:
        """Delete a member from the organization."""
        self.delete(pk=pf_user_id, path=f"pf_project/{pf_project_id}")

    def set_project_privilege(
        self, pf_user_id: UUID, pf_project_id: UUID, access_level: ProjectRole
    ) -> None:
        """Set a project-level role to a user."""
        self.partial_update(
            pk=pf_user_id,
            path=f"pf_project/{pf_project_id}/access_level",
            json={"access_level": access_level.value},
        )


class UserGroupClient(Client, UserRequestMixin):
    """User organization client."""

    def __init__(self, **kwargs):
        """Initialize user organization client."""
        self.initialize_user()
        super().__init__(pf_user_id=self.user_id, **kwargs)

    @property
    def url_path(self) -> Template:
        """Get an URL path."""
        return Template(self.url_provider.get_auth_uri("pf_user/$pf_user_id/pf_group"))

    def get_group_info(self) -> Dict[str, Any]:
        """Get organization info where user belongs to."""
        data = self.list(pagination=False)
        return data[0]


class UserGroupProjectClient(Client, UserRequestMixin, GroupRequestMixin):
    """User organization project client."""

    def __init__(self, **kwargs):
        """Initialize user organization project client."""
        self.initialize_user()
        self.initialize_group()
        super().__init__(pf_user_id=self.user_id, pf_group_id=self.group_id, **kwargs)

    @property
    def url_path(self) -> Template:
        """Get an URL path."""
        return Template(
            self.url_provider.get_auth_uri(
                "pf_user/$pf_user_id/pf_group/$pf_group_id/pf_project"
            )
        )

    def list_projects(self) -> List[Dict[str, Any]]:
        """List projects in the organization."""
        data = self.list(pagination=True)
        return data


class UserAccessKeyClient(Client, UserRequestMixin):
    """User access key client."""

    def __init__(self, **kwargs):
        """Initialize user access key client."""
        self.initialize_user()
        super().__init__(pf_user_id=self.user_id, **kwargs)

    @property
    def url_path(self) -> Template:
        """Get an URL path."""
        return Template(self.url_provider.get_auth_uri("pf_user"))

    def create_access_key(self, name: str) -> Dict[str, Any]:
        """Create a new access key."""
        data = self.post(path=f"{self.user_id}/api_key", json={"name": name})
        return data

    def delete_access_key(self, access_key_id: str) -> None:
        """Revoke an access key."""
        self.delete(pk=None, path=f"api_key/{access_key_id}")

    def list_access_keys(self) -> List[Dict[str, Any]]:
        """List access keys."""
        data = self.list(path=f"{self.user_id}/api_key", pagination=False)
        return data



# Copyright (c) 2022-present, FriendliAI Inc. All rights reserved.

"""PeriFlow GPT2 Checkpoint Converter."""

from __future__ import annotations

from typing import Any, Callable, Dict, List, cast

import numpy as np
import torch
from transformers import GPT2Config  # type: ignore[import]

from periflow.errors import CheckpointConversionError, NotSupportedCheckpointError
from periflow.logging import logger
from periflow.modules.converter.base import (
    DECODER_PREFIX,
    SUPPORTED_GELU_FAMILY,
    DecoderOnlyConverter,
)
from periflow.modules.converter.utils import (
    convert_tensor_to_np_array,
    get_tensor_from_state_dict,
    nontype_partial,
)


class GPT2LMHeadModelConverter(DecoderOnlyConverter):
    """GPT2LMHeadModel Architectures Converter Class."""

    def check_config(self) -> None:
        """Check if GPT2 architectures' config can be converted to PeriFlow format."""
        super().check_config()
        try:
            if (
                cast(GPT2Config, self.config).activation_function
                not in SUPPORTED_GELU_FAMILY
            ):
                raise NotSupportedCheckpointError(
                    invalid_option="'activation_function="
                    f"{cast(GPT2Config, self.config).activation_function}'",
                    valid_options=SUPPORTED_GELU_FAMILY,
                )
            if cast(GPT2Config, self.config).scale_attn_by_inverse_layer_idx:
                raise NotSupportedCheckpointError(
                    invalid_option="'scale_attn_by_inverse_layer_idx=True'",
                    valid_options=[False],
                )
            if not cast(GPT2Config, self.config).tie_word_embeddings:
                raise NotSupportedCheckpointError(
                    invalid_option="'tie_word_embeddings=False'",
                    valid_options=[True],
                )
            if cast(GPT2Config, self.config).layer_norm_epsilon != 1e-5:
                raise NotSupportedCheckpointError(
                    invalid_option="'layer_norm_epsilon="
                    f"{cast(GPT2Config, self.config).layer_norm_epsilon}'",
                    valid_options=[1e-5],
                )
        except AttributeError as exc:
            raise CheckpointConversionError(str(exc)) from exc

    def get_attributes(self) -> Dict[str, Any]:
        """Get checkpoint attributes."""
        config = cast(GPT2Config, self.config)

        logger.warn(
            "Since GPT2 uses absolute position embedding, 'max_length' cannot be "
            "larger than %d.",
            config.n_positions,
        )

        eos_token_id = self.get_eos_token_id()
        attr = {
            "model_type": self.model_type,
            "dtype": self.data_type.value,
            "head_size": self.decoder_head_size,
            "num_heads": self.decoder_num_attention_heads,
            "num_layers": self.decoder_layer_num,
            "max_length": config.n_positions,
            "vocab_size": config.vocab_size,
            "eos_token": eos_token_id if eos_token_id is not None else "FILL ME",
        }
        return attr

    @property
    def model_type(self) -> str:
        """Model type."""
        return "gpt"

    @property
    def non_transformer_convert_dict(
        self,
    ) -> Dict[str, Callable[[Dict[str, torch.Tensor], str], np.ndarray]]:
        """The convert_dict for non-transformer blocks in GPT2."""
        return {
            "wte/weight:0": nontype_partial(
                self.token_embed_weight_convert,
                per_layer_postfixes=["transformer.wte.weight"],
            ),
            DECODER_PREFIX
            + "/wpe/weight:0": nontype_partial(
                self.pos_embed_weight_convert,
                per_layer_postfixes=["transformer.wpe.weight"],
            ),
            DECODER_PREFIX
            + "/ln_f/gamma:0": nontype_partial(
                self.ln_weight_convert, per_layer_postfixes=["transformer.ln_f.weight"]
            ),
            DECODER_PREFIX
            + "/ln_f/beta:0": nontype_partial(
                self.ln_bias_convert, per_layer_postfixes=["transformer.ln_f.bias"]
            ),
        }

    def linear_weight_convert(
        self,
        state_dict: Dict[str, torch.Tensor],
        layer: str,
        per_layer_postfixes: List[str],
    ) -> np.ndarray:
        """Convert linear weight in GPT2, which does not need weight transpose."""
        assert len(per_layer_postfixes) == 1
        param = get_tensor_from_state_dict(state_dict, layer + per_layer_postfixes[0])
        return convert_tensor_to_np_array(param=param, data_type=self.data_type)

    @property
    def decoder_convert_dict(
        self,
    ) -> Dict[str, Callable[[Dict[str, torch.Tensor], str], np.ndarray]]:
        """The convert_dict for transformer blocks in GPT2."""
        return {
            "ln_1/gamma:0": nontype_partial(
                self.ln_weight_convert,
                per_layer_postfixes=[".ln_1.weight"],
            ),
            "ln_1/beta:0": nontype_partial(
                self.ln_bias_convert,
                per_layer_postfixes=[".ln_1.bias"],
            ),
            "attn/c_attn/bias:0": nontype_partial(
                self.linear_bias_convert,
                per_layer_postfixes=[".attn.c_attn.bias"],
            ),
            "attn/c_proj/bias:0": nontype_partial(
                self.linear_bias_convert,
                per_layer_postfixes=[".attn.c_proj.bias"],
            ),
            "ln_2/gamma:0": nontype_partial(
                self.ln_weight_convert,
                per_layer_postfixes=[".ln_2.weight"],
            ),
            "ln_2/beta:0": nontype_partial(
                self.ln_bias_convert,
                per_layer_postfixes=[".ln_2.bias"],
            ),
            "mlp/c_fc/bias:0": nontype_partial(
                self.linear_bias_convert,
                per_layer_postfixes=[".mlp.c_fc.bias"],
            ),
            "mlp/c_proj/bias:0": nontype_partial(
                self.linear_bias_convert,
                per_layer_postfixes=[".mlp.c_proj.bias"],
            ),
            "attn/c_attn/weight:0": nontype_partial(
                self.linear_weight_convert,
                per_layer_postfixes=[".attn.c_attn.weight"],
            ),
            "attn/c_proj/weight:0": nontype_partial(
                self.linear_weight_convert,
                per_layer_postfixes=[".attn.c_proj.weight"],
            ),
            "mlp/c_fc/weight:0": nontype_partial(
                self.linear_weight_convert,
                per_layer_postfixes=[".mlp.c_fc.weight"],
            ),
            "mlp/c_proj/weight:0": nontype_partial(
                self.linear_weight_convert,
                per_layer_postfixes=[".mlp.c_proj.weight"],
            ),
        }

    @property
    def decoder_layer_prefix(self) -> str:
        """The layer name prefix used before GPT2's transformer block number."""
        return "transformer.h."

    @property
    def decoder_layer_num(self) -> int:
        """The number of decoder layers in GPT2."""
        return cast(GPT2Config, self.config).num_hidden_layers

    @property
    def decoder_hidden_size(self) -> int:
        """The hidden size in GPT2."""
        return cast(GPT2Config, self.config).hidden_size

    @property
    def decoder_num_attention_heads(self) -> int:
        """The number of attention heads in GPT2."""
        return cast(GPT2Config, self.config).num_attention_heads

    @property
    def decoder_num_kv_attention_heads(self) -> int:
        """The number of key-value attention heads in gpt2."""
        return self.decoder_num_attention_heads

    @property
    def decoder_head_size(self) -> int:
        """The head siez of GPT2."""
        return self.decoder_hidden_size // self.decoder_num_attention_heads



# Copyright (c) 2022-present, FriendliAI Inc. All rights reserved.

"""PeriFlow MPTForCausalLM QuantizerHook."""

# mypy: ignore-errors

from __future__ import annotations

from typing import Any, Dict, Iterator, List, Tuple, Type

import torch

from periflow.modules.quantizer.awq.base import AWQHook
from periflow.modules.quantizer.schema.data import ModuleName, QuantInput, TFQuantInputs


class AWQMPTHook(AWQHook):
    """AWQ Hook for MPTForCausalLM."""

    def add_pre_scaler(self, model: torch.nn.Module) -> torch.nn.Module:
        """Adds scaler to MPTForCausalLM."""
        for tf_block in self.get_tf_blocks(model):
            attn_fc_scaler = self._register_pre_scaler(
                tf_block.attn.out_proj,
            )
            tf_block.attn.add_module("scaler", attn_fc_scaler)
            ff2_scaler = self._register_pre_scaler(tf_block.ffn.down_proj)
            tf_block.ffn.add_module("scaler", ff2_scaler)
        return model

    def get_inspect_module_types(
        self, block: torch.nn.Module
    ) -> Tuple[Type[torch.nn.Module], ...]:
        """Returns the type of linear layer (etc. qkv, linear layer) in transformer block."""
        return (type(block.attn),)

    def iter_inspect_modules(
        self,
        block: torch.nn.Module,
        intra_kwargs: Dict[str, Any],
    ) -> Iterator[
        Tuple[
            List[torch.nn.Module],
            List[torch.nn.Module],
            ModuleName,
            torch.nn.Module,
            Dict[str, Any],
        ]
    ]:
        """Returns iterator of layers in modules."""
        # qkv proj
        yield (
            [block.norm_1],
            [block.attn.Wqkv],
            "attn.Wqkv",
            block.attn,
            intra_kwargs["attn.Wqkv"],
        )
        # attn out proj
        yield (
            [block.attn.scaler],
            [block.attn.out_proj],
            "attn.out_proj",
            block.attn.out_proj,
            {},
        )
        # ff1
        yield (
            [block.norm_2],
            [block.ffn.up_proj],
            "ffn.up_proj",
            block.ffn,
            {},
        )
        # ff2
        yield (
            [block.ffn.scaler],
            [block.ffn.down_proj],
            "ffn.down_proj",
            block.ffn.down_proj,
            {},
        )

    def iter_quant_inputs(self, model: torch.nn.Module) -> Iterator[TFQuantInputs]:
        """Returns the layers which should be quantized in transformer block of MPTForCausalLM."""
        for index, decoder_layer in enumerate(
            self.get_tf_blocks(model)  # type: ignore[union-attr, arg-type]
        ):
            self_attn = decoder_layer.attn
            attn_weight_outdim = self_attn.Wqkv.weight.size(0)  # OutDim
            fc1 = decoder_layer.ffn.up_proj  # type: ignore
            fc2 = decoder_layer.ffn.down_proj  # type: ignore

            yield TFQuantInputs(
                layer_index=index,
                q=QuantInput(
                    self_attn.Wqkv.weight,  # type: ignore
                    f"{self.quantized_layer_prefix}.{index}.attn.Wqkv",
                    0,
                    attn_weight_outdim // 3,
                ),
                k=QuantInput(
                    self_attn.Wqkv.weight,  # type: ignore
                    f"{self.quantized_layer_prefix}.{index}.attn.Wqkv",
                    attn_weight_outdim // 3,
                    attn_weight_outdim // 3 * 2,
                ),
                v=QuantInput(
                    self_attn.Wqkv.weight,  # type: ignore
                    f"{self.quantized_layer_prefix}.{index}.attn.Wqkv",
                    attn_weight_outdim // 3 * 2,
                    attn_weight_outdim,
                ),
                attn_fc=QuantInput(
                    self_attn.out_proj.weight,  # type: ignore
                    f"{self.quantized_layer_prefix}.{index}.attn.out_proj",
                    None,
                    None,
                ),
                ff1=QuantInput(
                    fc1.weight,  # type: ignore
                    f"{self.quantized_layer_prefix}.{index}.ffn.up_proj",
                    None,
                    None,
                ),
                ff2=QuantInput(
                    fc2.weight,  # type: ignore
                    f"{self.quantized_layer_prefix}.{index}.ffn.down_proj",
                    None,
                    None,
                ),
            )

    def get_linear_layer_types(self) -> Tuple[Type[torch.nn.Module]]:
        """Returns the linear layer types in MPTForCausalLM."""
        return (torch.nn.Linear,)

    def get_tf_blocks(self, model: torch.nn.Module) -> List[torch.nn.Module]:
        """Returns the transformer blocks in MPTForCausalLM."""
        return model.transformer.blocks  # type: ignore

    @property
    def quantized_layer_prefix(self) -> str:
        """Returns the prefix of the transformer block name."""
        return "transformer.blocks"

    @property
    def avoid_clipping_layer_names(self) -> List[str]:
        """Returns the layer names which should be avoided for AWQ clipping."""
        return ["Wqkv"]



# Copyright (c) 2022-present, FriendliAI Inc. All rights reserved.

"""PeriFlow LlamaForCausalLM QuantizerHook."""

# mypy: ignore-errors

from __future__ import annotations

from dataclasses import dataclass
from typing import Any, Callable, Dict, Iterator, List, Tuple, Type, cast

import numpy as np
import torch

from periflow.modules.converter.utils import convert_to_gpt_j_params, nontype_partial
from periflow.modules.quantizer.awq.base import AWQHook
from periflow.modules.quantizer.schema.config import AWQConfig
from periflow.modules.quantizer.schema.data import (
    ModuleName,
    QuantInput,
    TFQuantInputs,
    TFQuantResults,
    WeightOnlyQuantResult,
)
from periflow.modules.quantizer.utils import (
    get_weight_only_quant_scales,
    quantized_linear_weight_convert,
    scale_convert,
)


@dataclass
class LlamaTFQuantInputs(TFQuantInputs):
    """Dataclass for quantization input per layer in LlamaForCausalLM."""

    ff_gate: QuantInput


@dataclass
class LlamaTFQuantResults(TFQuantResults):
    """Dataclass for quantization result per layer in LlamaForCausalLM."""

    ff_gate: WeightOnlyQuantResult


class AWQLlamaHook(AWQHook):
    """AWQ Hook for LlamaForCausalLM."""

    def __init__(self, quant_config, converter):
        """Initialize AWQLlamaHook."""
        super().__init__(quant_config, converter)
        config = converter.config
        self.data_type = converter.data_type
        self.num_attention_heads = config.num_attention_heads
        if config.num_key_value_heads is None:
            self.num_kv_attention_heads = self.num_attention_heads
        else:
            self.num_kv_attention_heads = config.num_key_value_heads
        self.hidden_size = config.hidden_size
        self.head_size = self.hidden_size // self.num_attention_heads
        self.rotary_dim = self.head_size
        self.require_attn_fc_pre_scale = (
            self.num_attention_heads != self.num_kv_attention_heads
        )

    def add_pre_scaler(self, model: torch.nn.Module) -> torch.nn.Module:
        """Adds scaler to LlamaForCausalLM."""
        for tf_block in self.get_tf_blocks(model):
            if self.require_attn_fc_pre_scale:
                attn_fc_scaler = self._register_pre_scaler(
                    tf_block.self_attn.o_proj,
                )
                tf_block.self_attn.add_module("scaler", attn_fc_scaler)
        return model

    def get_inspect_module_types(
        self, block: torch.nn.Module
    ) -> Tuple[type[torch.nn.Module], ...]:
        """Returns the layer types in inspected blocks."""
        return (type(block.self_attn),)

    def iter_inspect_modules(
        self,
        block: torch.nn.Module,
        intra_kwargs: Dict[str, Any],
    ) -> Iterator[
        Tuple[
            List[torch.nn.Module],
            List[torch.nn.Module],
            ModuleName,
            torch.nn.Module,
            Dict[str, Any],
        ]
    ]:
        """Returns iterator of layers in blocks."""
        # qkv proj
        yield (
            [block.input_layernorm],
            [
                block.self_attn.q_proj,
                block.self_attn.k_proj,
                block.self_attn.v_proj,
            ],
            "self_attn.q_proj",  # "self_attn.k_proj" or "self_attn.v_proj" is also valid
            block.self_attn,
            intra_kwargs["self_attn"],
        )
        # attn out proj
        yield (
            [
                block.self_attn.scaler
                if self.require_attn_fc_pre_scale
                else block.self_attn.v_proj
            ],
            [block.self_attn.o_proj],
            "self_attn.o_proj",
            block.self_attn.o_proj,
            {},
        )
        # ff1
        yield (
            [block.post_attention_layernorm],
            [
                block.mlp.up_proj,
                block.mlp.gate_proj,
            ],
            "mlp.gate_proj",  # "mlp.up_proj" is also valid
            block.mlp,
            {},
        )
        # ff2
        yield (
            [block.mlp.up_proj],
            [block.mlp.down_proj],
            "mlp.down_proj",
            block.mlp.down_proj,
            {},
        )

    def reshape_qkv_weight(
        self, attn_layer: torch.nn.Module
    ) -> Tuple[torch.Tensor, torch.Tensor, torch.Tensor]:
        """Reshape LlamaForCausalLM's qkv weight."""
        q_weight, k_weight, v_weight = (
            cast(torch.nn.Linear, attn_layer.q_proj).weight,
            cast(torch.nn.Linear, attn_layer.k_proj).weight,
            cast(torch.nn.Linear, attn_layer.v_proj).weight,
        )
        q_weight = q_weight.reshape(
            self.num_attention_heads,
            self.head_size,
            self.hidden_size,
        )
        k_weight = k_weight.reshape(
            self.num_kv_attention_heads,
            self.head_size,
            self.hidden_size,
        )
        q_weight = convert_to_gpt_j_params(q_weight, self.rotary_dim)
        k_weight = convert_to_gpt_j_params(k_weight, self.rotary_dim)
        q_weight = q_weight.reshape(
            self.num_attention_heads * self.head_size,
            self.hidden_size,
        )
        k_weight = k_weight.reshape(
            self.num_kv_attention_heads * self.head_size,
            self.hidden_size,
        )
        return q_weight, k_weight, v_weight

    def iter_quant_inputs(self, model: torch.nn.Module) -> Iterator[TFQuantInputs]:
        """Returns the layers which should be quantized in transformer block of LlamaForCausalLM."""
        for index, decoder_layer in enumerate(
            self.get_tf_blocks(model)  # type: ignore[union-attr, arg-type]
        ):
            self_attn = decoder_layer.self_attn
            q_weight, k_weight, v_weight = self.reshape_qkv_weight(self_attn)
            fc1 = decoder_layer.mlp.up_proj
            ff_gate = decoder_layer.mlp.gate_proj
            fc2 = decoder_layer.mlp.down_proj

            yield LlamaTFQuantInputs(
                layer_index=index,
                q=QuantInput(
                    q_weight,
                    f"{self.quantized_layer_prefix}.{index}.self_attn.q_proj",
                    None,
                    None,
                ),
                k=QuantInput(
                    k_weight,
                    f"{self.quantized_layer_prefix}.{index}.self_attn.k_proj",
                    None,
                    None,
                ),
                v=QuantInput(
                    v_weight,
                    f"{self.quantized_layer_prefix}.{index}.self_attn.v_proj",
                    None,
                    None,
                ),
                attn_fc=QuantInput(
                    self_attn.o_proj.weight,
                    f"{self.quantized_layer_prefix}.{index}.self_attn.o_proj",
                    None,
                    None,
                ),
                ff1=QuantInput(
                    fc1.weight,
                    f"{self.quantized_layer_prefix}.{index}.mlp.up_proj",
                    None,
                    None,
                ),
                ff_gate=QuantInput(
                    ff_gate.weight,
                    f"{self.quantized_layer_prefix}.{index}.mlp.gate_proj",
                    None,
                    None,
                ),
                ff2=QuantInput(
                    fc2.weight,
                    f"{self.quantized_layer_prefix}.{index}.mlp.down_proj",
                    None,
                    None,
                ),
            )

    def get_quant_result(
        self,
        quant_input: TFQuantInputs,
        **kwargs: Any,
    ) -> TFQuantResults:
        """Get quantization result for a specific layer in LlamaForCausalLM."""
        awq_args = cast(AWQConfig, self.quant_config).awq_args

        def get_scale(quant_input: QuantInput) -> WeightOnlyQuantResult:
            weight, name, start, end = (
                quant_input.weight,
                quant_input.name,
                quant_input.start_offset,
                quant_input.end_offset,
            )

            return get_weight_only_quant_scales(
                layer_name=name,
                w=weight[start:end],
                q_bit=awq_args.quant_bit,
                q_group_size=awq_args.quant_group_size,
            )

        quant_input = cast(LlamaTFQuantInputs, quant_input)
        return LlamaTFQuantResults(
            layer_prefix_with_index=f"{self.quantized_layer_prefix}.{quant_input.layer_index}",
            q=get_scale(quant_input.q),
            k=get_scale(quant_input.k),
            v=get_scale(quant_input.v),
            attn_fc=get_scale(quant_input.attn_fc),
            ff1=get_scale(quant_input.ff1),
            ff_gate=get_scale(quant_input.ff_gate),
            ff2=get_scale(quant_input.ff2),
        )

    def get_linear_layer_types(self) -> Tuple[Type[torch.nn.Module]]:
        """Returns the linear layer types in LlamaForCausalLM."""
        return (torch.nn.Linear,)

    def get_tf_blocks(self, model: torch.nn.Module) -> List[torch.nn.Module]:
        """Returns the transformer blocks in LlamaForCausalLM."""
        return model.model.layers

    @property
    def quantized_param_names(self) -> List[str]:
        """Returns the parameter names in LlamaForCausalLM."""
        return super().quantized_param_names + [
            "mlp/c_gate/weight:0",
        ]

    @property
    def modified_layers_convert_dict(
        self,
    ) -> Dict[str, Callable[[Dict[str, torch.Tensor], str], np.ndarray]]:
        """Return the convert_dict for modified layers."""
        return (
            {
                "attn/c_proj/awq/pre_scale:0": nontype_partial(
                    scale_convert,
                    per_layer_postfixes=[".self_attn.scaler.scale"],
                    data_type="fp32",
                ),
            }
            if self.require_attn_fc_pre_scale
            else {}
        )

    @property
    def avoid_clipping_layer_names(self) -> List[str]:
        """Returns the layer names which should avoid AWQ clipping."""
        return ["q_proj", "k_proj", "v_proj"]

    @property
    def quantized_convert_dict(
        self,
    ) -> Dict[str, Callable[[Dict[str, torch.Tensor], str], np.ndarray]]:
        """Return the convert_dict for quantized layers."""
        n_bit = cast(AWQConfig, self.quant_config).awq_args.quant_bit
        convert_dict = super().quantized_convert_dict
        convert_dict.update(
            {
                "mlp/c_gate/awq/scale:0": nontype_partial(
                    scale_convert,
                    per_layer_postfixes=[".ff_gate.woq_weight_scale"],
                    data_type=self.data_type,
                ),
                "mlp/c_gate/awq/zero:0": nontype_partial(
                    scale_convert,
                    per_layer_postfixes=[".ff_gate.woq_weight_zp"],
                    data_type=self.data_type,
                ),
                "mlp/c_gate/awq/weight:0": nontype_partial(
                    quantized_linear_weight_convert,
                    per_layer_postfixes=[".ff_gate.woq_weight"],
                    n_bit=n_bit,
                ),
            }
        )
        return convert_dict



# Copyright (c) 2022-present, FriendliAI Inc. All rights reserved.

"""PeriFlow Quantizer Base."""

from __future__ import annotations

import os
from abc import ABC, abstractmethod
from typing import Any, Callable, Dict, Iterator, List, Tuple, Type

import datasets  # type: ignore[import]
import h5py  # type: ignore[import]
import numpy as np
import torch
from tqdm.auto import tqdm

from periflow.enums import (
    QuantDatasetFormat,  # TODO: move this to periflow/modules/converter/enums.py
)
from periflow.errors import NotSupportedQuantConfigError
from periflow.modules.converter.base import DECODER_PREFIX, OneOfConverter
from periflow.modules.converter.interface import ModelConversionInterface
from periflow.modules.quantizer.schema.config import CommonQuantConfig
from periflow.modules.quantizer.schema.data import TFQuantInputs, TFQuantResults


class AbstractQuantHook(ABC):
    """Quantization Hook for a specific model architecture."""

    def __init__(self, quant_config: Dict[str, Any], converter: OneOfConverter):
        """Initialize the Quantization Hook.

        Args:
            quant_config: Quantization configuration.
            converter (OneOfConverter): Converter for a specific model architecture.
        """
        self.quant_config = quant_config
        self.converter = converter

    @abstractmethod
    def get_tf_blocks(self, model: torch.nn.Module) -> List[torch.nn.Module]:
        """Returns the transformer blocks."""

    @abstractmethod
    def get_linear_layer_types(self) -> Tuple[Type[torch.nn.Module]]:
        """Returns the type of linear layer (etc. qkv, linear layer) in transformer block."""

    @abstractmethod
    def iter_quant_inputs(self, model: torch.nn.Module) -> Iterator[TFQuantInputs]:
        """Returns the layers which should be quantized in transformer blocks."""

    @abstractmethod
    def get_quant_result(
        self,
        quant_input: TFQuantInputs,
        **kwargs: Any,
    ) -> TFQuantResults:
        """Returns the quantization result of the layer."""

    @property
    @abstractmethod
    def quantized_convert_dict(
        self,
    ) -> Dict[str, Callable[[Dict[str, torch.Tensor], str], np.ndarray]]:
        """Return the convert_dict for quantized layers."""

    @property
    @abstractmethod
    def modified_layers_convert_dict(
        self,
    ) -> Dict[str, Callable[[Dict[str, torch.Tensor], str], np.ndarray]]:
        """Return the convert_dict for modified layers."""

    @property
    @abstractmethod
    def quantized_layer_prefix(self) -> str:
        """Returns the prefix of the transformer block name."""
        return self.converter.decoder_layer_prefix

    @property
    def quantized_param_names(self) -> List[str]:
        """Return the parameter names of quantized layers."""
        return [
            "attn/c_attn/weight:0",
            "attn/c_proj/weight:0",
            "mlp/c_fc/weight:0",
            "mlp/c_proj/weight:0",
        ]


class AbstractQuantizer(ABC):
    """Abstract Quantizer for a specific model architecture."""

    def __init__(
        self,
        hook: AbstractQuantHook,
        config: CommonQuantConfig,
        converter: OneOfConverter,
    ):
        """Initialize the Quantizer.

        Args:
            hook (AbstractQuantHook): Quantization Hook for a specific model architecture
            config (CommonQuantConfig): Quantization configuration.
            converter (OneOfConverter): Converter for a specific model architecture.

        """
        self.hook = hook
        self.quant_config = config
        self.converter = converter

    @abstractmethod
    def get_calib_dataset(
        self,
    ) -> datasets.Dataset:
        """Get calibration dataset."""

    @abstractmethod
    def pre_quantize(
        self,
        model: torch.nn.Module,
    ) -> None:
        """Pre-procedure that should be called before quantize() is called."""

    @abstractmethod
    def quantize(
        self,
        model: torch.nn.Module,
    ) -> Iterator[TFQuantResults]:
        """Setting Quantizer from config and Quantize model."""

    @abstractmethod
    def get_quantized_state_dict(
        self,
        model: torch.nn.Module,
        quant_result_iter: Iterator[TFQuantResults],
    ) -> Dict[str, torch.Tensor]:
        """Get quantized state dict."""


class CommonQuantizer(AbstractQuantizer, ModelConversionInterface):
    """Common Quantizer."""

    def check_config(self) -> None:
        """Check if the quantization config is valid."""
        calibration_dataset_config = self.quant_config.calibration_dataset
        data_path_or_name = calibration_dataset_config.path_or_name
        percentile = self.quant_config.percentile
        if percentile <= 0 or percentile > 100:
            raise NotSupportedQuantConfigError(
                invalid_option=str(percentile),
                valid_options=["0 < percentile <= 100"],
            )
        if not os.path.exists(data_path_or_name):
            data_name = data_path_or_name.split(":")[0]
            if data_name not in datasets.list_datasets():
                raise NotSupportedQuantConfigError(
                    invalid_option=data_name,
                    valid_options=[datasets.list_datasets(), "local path"],
                )
        else:
            if calibration_dataset_config.format not in QuantDatasetFormat:
                raise NotSupportedQuantConfigError(
                    invalid_option=calibration_dataset_config.format,
                    valid_options=list(QuantDatasetFormat),
                )
        try:
            torch.device(self.quant_config.device)
        except ValueError as err:
            raise NotSupportedQuantConfigError(
                invalid_option=self.quant_config.device,
                valid_options=["cpu", "cuda"],
            ) from err

    def get_convert_dict(
        self,
    ) -> Dict[str, Dict[str, Callable[[Dict[str, torch.Tensor], str], np.ndarray]]]:
        """Return the convert_dict for the model."""
        convert_dict = self.converter.get_convert_dict()
        for param_name in self.hook.quantized_param_names:
            del convert_dict["decoder"][param_name]

        convert_dict["decoder"].update(self.hook.quantized_convert_dict)
        convert_dict["decoder"].update(self.hook.modified_layers_convert_dict)
        return convert_dict

    def get_attributes(self) -> Dict[str, Any]:
        """Return the attributes of the converted model."""
        return self.converter.get_attributes()

    def convert(
        self,
        model: torch.nn.Module,
        output_path: str,
        convert_dict: Dict[
            str, Dict[str, Callable[[Dict[str, torch.Tensor], str], np.ndarray]]
        ],
    ) -> None:
        """Convert Huggingface Model to PeriFlow format(.h5).

        Args:
            model (torch.nn.Module): Huggingface model.
            output_path (str): Path to save the converted model.
            convert_dict (Dict[Callable[[Dict[str, torch.Tensor], str], np.ndarray]]):
                Dictionary of mapping converted params name to conversion functions.

        """
        self.pre_quantize(model)
        quant_result_iter = self.quantize(model)
        state_dict = model.state_dict()
        state_dict.update(self.get_quantized_state_dict(model, quant_result_iter))

        total_layers = len(
            convert_dict["decoder"]
        ) * self.converter.decoder_layer_num + len(convert_dict["non-transformer"])
        with h5py.File(output_path, "w") as out_f, tqdm(
            total=total_layers, desc="Converting", unit="tensor"
        ) as pbar:
            self.converter.convert_decoder_layers(
                state_dict=state_dict,
                convert_dict=convert_dict["decoder"],
                out_f=out_f.create_group(DECODER_PREFIX),
                pbar=pbar,
            )
            self.converter.convert_non_transformer_layers(
                state_dict=state_dict,
                convert_dict=convert_dict["non-transformer"],
                out_f=out_f,
                pbar=pbar,
            )



# Copyright (c) 2022-present, FriendliAI Inc. All rights reserved.

"""PeriFlow Bloom Checkpoint Converter."""

from __future__ import annotations

from typing import Any, Callable, Dict, List, cast

import numpy as np
import torch
from transformers import BloomConfig  # type: ignore[import]

from periflow.errors import CheckpointConversionError, NotSupportedCheckpointError
from periflow.logging import logger
from periflow.modules.converter.base import DECODER_PREFIX, DecoderOnlyConverter
from periflow.modules.converter.utils import (
    convert_tensor_to_np_array,
    get_tensor_from_state_dict,
    nontype_partial,
)


class BloomForCausalLMConverter(DecoderOnlyConverter):
    """BloomForCausalLM Architectures Converter Class."""

    def check_config(self) -> None:
        """Check if Bloom architectures' config can be converted to PeriFlow format."""
        super().check_config()
        try:
            if cast(BloomConfig, self.config).apply_residual_connection_post_layernorm:
                raise NotSupportedCheckpointError(
                    invalid_option="apply_residual_connection_post_layernorm=True",
                    valid_options=[False],
                )
            if cast(BloomConfig, self.config).slow_but_exact:
                raise NotSupportedCheckpointError(
                    invalid_option="slow_but_exact=True", valid_options=[False]
                )
            if not cast(BloomConfig, self.config).tie_word_embeddings:
                raise NotSupportedCheckpointError(
                    invalid_option="tie_word_embeddings=False", valid_options=[True]
                )
            if cast(BloomConfig, self.config).layer_norm_epsilon != 1e-5:
                raise NotSupportedCheckpointError(
                    invalid_option="layer_norm_epsilon="
                    f"{cast(BloomConfig, self.config).layer_norm_epsilon}",
                    valid_options=[1e-5],
                )
        except AttributeError as exc:
            raise CheckpointConversionError(str(exc)) from exc

    def qkv_weight_convert(
        self,
        state_dict: Dict[str, torch.Tensor],
        layer: str,
        per_layer_postfixes: List[str],
    ) -> np.ndarray:
        """qkv_weight_convert for Bloom's attention layer."""
        assert len(per_layer_postfixes) == 1
        qkv_weight = get_tensor_from_state_dict(
            state_dict, layer + per_layer_postfixes[0]
        )
        split_qkv_weight_list = torch.split(qkv_weight, self.decoder_head_size, dim=0)
        qkv_weight_list = [
            torch.cat(
                [
                    split_qkv_weight_list[j * 3 + i]
                    for j in range(self.decoder_num_attention_heads)
                ],
                dim=0,
            ).reshape(-1, self.decoder_hidden_size)
            for i in range(3)
        ]

        qkv_weight = torch.cat(qkv_weight_list, dim=0).transpose(0, 1)
        return convert_tensor_to_np_array(qkv_weight, self.data_type)

    def qkv_bias_convert(
        self,
        state_dict: Dict[str, torch.Tensor],
        layer: str,
        per_layer_postfixes: str,  # type: ignore[override]
    ) -> np.ndarray:
        """qkv_bias_convert for Bloom's attention layer."""
        assert len(per_layer_postfixes) == 1
        qkv_bias = get_tensor_from_state_dict(
            state_dict, layer + per_layer_postfixes[0]
        )
        split_qkv_bias_list = torch.split(qkv_bias, self.decoder_head_size, dim=0)
        qkv_bias_list = [
            torch.cat(
                [
                    split_qkv_bias_list[j * 3 + i]
                    for j in range(self.decoder_num_attention_heads)
                ],
                dim=0,
            )
            for i in range(3)
        ]

        qkv_bias = torch.cat(qkv_bias_list, dim=0)
        return convert_tensor_to_np_array(qkv_bias, self.data_type)

    def get_attributes(self) -> Dict[str, Any]:
        """Get checkpoint attributes."""
        config = cast(BloomConfig, self.config)

        logger.warn(
            "The 'max_length' field is left blank as it cannot be automatically configured. "
            "You must determine the 'max_length' according to your needs. The Bloom model does "
            "not rely on absolute position embeddings, allowing you to choose any "
            "suitable value."
        )

        eos_token_id = self.get_eos_token_id()
        attr = {
            "model_type": self.model_type,
            "dtype": self.data_type.value,
            "head_size": self.decoder_head_size,
            "num_heads": self.decoder_num_attention_heads,
            "num_layers": self.decoder_layer_num,
            "max_length": "FILL ME",
            "vocab_size": config.vocab_size,
            "eos_token": eos_token_id if eos_token_id is not None else "FILL ME",
        }
        return attr

    @property
    def model_type(self) -> str:
        """Model type."""
        return "bloom"

    @property
    def non_transformer_convert_dict(
        self,
    ) -> Dict[str, Callable[[Dict[str, torch.Tensor], str], np.ndarray]]:
        """The convert_dict for non-transformer blocks in Bloom."""
        return {
            "wte/weight:0": nontype_partial(
                self.token_embed_weight_convert,
                per_layer_postfixes=["transformer.word_embeddings.weight"],
            ),
            "wte/ln/gamma:0": nontype_partial(
                self.ln_weight_convert,
                per_layer_postfixes=["transformer.word_embeddings_layernorm.weight"],
            ),
            "wte/ln/beta:0": nontype_partial(
                self.ln_bias_convert,
                per_layer_postfixes=["transformer.word_embeddings_layernorm.bias"],
            ),
            DECODER_PREFIX
            + "/ln_f/gamma:0": nontype_partial(
                self.ln_weight_convert, per_layer_postfixes=["transformer.ln_f.weight"]
            ),
            DECODER_PREFIX
            + "/ln_f/beta:0": nontype_partial(
                self.ln_bias_convert, per_layer_postfixes=["transformer.ln_f.bias"]
            ),
        }

    @property
    def decoder_convert_dict(
        self,
    ) -> Dict[str, Callable[[Dict[str, torch.Tensor], str], np.ndarray]]:
        """The convert_dict for transformer blocks in Bloom."""
        return {
            "ln_1/gamma:0": nontype_partial(
                self.ln_weight_convert,
                per_layer_postfixes=[".input_layernorm.weight"],
            ),
            "ln_1/beta:0": nontype_partial(
                self.ln_bias_convert,
                per_layer_postfixes=[".input_layernorm.bias"],
            ),
            "attn/c_attn/bias:0": nontype_partial(
                self.qkv_bias_convert,
                per_layer_postfixes=[".self_attention.query_key_value.bias"],
            ),
            "attn/c_proj/bias:0": nontype_partial(
                self.linear_bias_convert,
                per_layer_postfixes=[".self_attention.dense.bias"],
            ),
            "ln_2/gamma:0": nontype_partial(
                self.ln_weight_convert,
                per_layer_postfixes=[".post_attention_layernorm.weight"],
            ),
            "ln_2/beta:0": nontype_partial(
                self.ln_bias_convert,
                per_layer_postfixes=[".post_attention_layernorm.bias"],
            ),
            "mlp/c_fc/bias:0": nontype_partial(
                self.linear_bias_convert,
                per_layer_postfixes=[".mlp.dense_h_to_4h.bias"],
            ),
            "mlp/c_proj/bias:0": nontype_partial(
                self.linear_bias_convert,
                per_layer_postfixes=[".mlp.dense_4h_to_h.bias"],
            ),
            "attn/c_attn/weight:0": nontype_partial(
                self.qkv_weight_convert,
                per_layer_postfixes=[".self_attention.query_key_value.weight"],
            ),
            "attn/c_proj/weight:0": nontype_partial(
                self.linear_weight_convert,
                per_layer_postfixes=[".self_attention.dense.weight"],
            ),
            "mlp/c_fc/weight:0": nontype_partial(
                self.linear_weight_convert,
                per_layer_postfixes=[".mlp.dense_h_to_4h.weight"],
            ),
            "mlp/c_proj/weight:0": nontype_partial(
                self.linear_weight_convert,
                per_layer_postfixes=[".mlp.dense_4h_to_h.weight"],
            ),
        }

    @property
    def decoder_layer_prefix(self) -> str:
        """The layer name prefix used before Bloom's transformer block number."""
        return "transformer.h."

    @property
    def decoder_layer_num(self) -> int:
        """The number of decoder layers in Bloom."""
        return cast(BloomConfig, self.config).num_hidden_layers

    @property
    def decoder_hidden_size(self) -> int:
        """Return the hidden size in Bloom."""
        return cast(BloomConfig, self.config).hidden_size

    @property
    def decoder_num_attention_heads(self) -> int:
        """The number of attention heads in Bloom."""
        return cast(BloomConfig, self.config).num_attention_heads

    @property
    def decoder_num_kv_attention_heads(self) -> int:
        """The number of key-value attention heads in bloom."""
        return self.decoder_num_attention_heads

    @property
    def decoder_head_size(self) -> int:
        """The size of each attention head in Bloom."""
        return self.decoder_hidden_size // self.decoder_num_attention_heads



# Copyright (c) 2022-present, FriendliAI Inc. All rights reserved.

# pylint: disable=too-many-arguments

"""PeriFlow Checkpoint Clients."""

from __future__ import annotations

from string import Template
from typing import Any, Dict, List, Optional
from uuid import UUID

from periflow.client.base import (
    Client,
    GroupRequestMixin,
    ProjectRequestMixin,
    UploadableClient,
    UserRequestMixin,
)
from periflow.enums import CheckpointCategory, StorageType


class CheckpointClient(Client[UUID]):
    """Checkpoint client."""

    @property
    def url_path(self) -> Template:
        """Get an URL path."""
        return Template(self.url_provider.get_mr_uri("models/"))

    def get_checkpoint(self, checkpoint_id: UUID) -> Dict[str, Any]:
        """Get a checkpoint info."""
        data = self.retrieve(pk=checkpoint_id)
        return data

    def get_first_checkpoint_form(self, checkpoint_id: UUID) -> UUID:
        """Get the first form of the checkpoint."""
        data = self.retrieve(pk=checkpoint_id)
        return UUID(data["forms"][0]["id"])

    def activate_checkpoint(self, checkpoint_id: UUID) -> Dict[str, Any]:
        """Make checkpoint status active."""
        data = self.partial_update(pk=checkpoint_id, json={"status": "Active"})
        return data

    def delete_checkpoint(self, checkpoint_id: UUID) -> None:
        """Delete a checkpoint."""
        self.delete(pk=checkpoint_id)

    def restore_checkpoint(self, checkpoint_id: UUID) -> Dict[str, Any]:
        """Restore a soft-deleted checkpoint."""
        data = self.post(path=f"{checkpoint_id}/restore/")
        return data


class CheckpointFormClient(UploadableClient[UUID]):
    """Checkpoint form client."""

    @property
    def url_path(self) -> Template:
        """Get an URL path."""
        return Template(self.url_provider.get_mr_uri("model_forms/"))

    def update_checkpoint_files(
        self,
        ckpt_form_id: UUID,
        files: List[Dict[str, Any]],
    ) -> Dict[str, Any]:
        """Update checkpoint file metadata."""
        data = self.partial_update(pk=ckpt_form_id, json={"files": files})
        return data

    def get_checkpoint_download_urls(self, ckpt_form_id: UUID) -> List[Dict[str, Any]]:
        """Get presigned URLs to download a model checkpoint."""
        data = self.retrieve(pk=ckpt_form_id, path="download/")
        return data["files"]


class GroupProjectCheckpointClient(
    Client, UserRequestMixin, GroupRequestMixin, ProjectRequestMixin
):
    """Organization project checkpoint client."""

    def __init__(self, **kwargs):
        """Initialize organization project checkpoint client."""
        self.initialize_user()
        self.initialize_group()
        self.initialize_project()
        super().__init__(group_id=self.group_id, project_id=self.project_id, **kwargs)

    @property
    def url_path(self) -> Template:
        """Get an URL path."""
        return Template(
            self.url_provider.get_mr_uri("orgs/$group_id/prjs/$project_id/models/")
        )

    def list_checkpoints(
        self, category: Optional[CheckpointCategory], limit: int, deleted: bool
    ) -> List[Dict[str, Any]]:
        """List checkpoints."""
        params = {}
        if category is not None:
            params["category"] = category.value
        if deleted:
            params["status"] = "deleted"

        checkpoints = self.list(
            pagination=True,
            limit=limit,
            params=params,
        )
        return checkpoints

    def create_checkpoint(
        self,
        name: str,
        vendor: StorageType,
        region: str,
        credential_id: Optional[UUID],
        iteration: Optional[int],
        storage_name: str,
        files: List[Dict[str, Any]],
        dist_config: Dict[str, Any],
        attributes: Dict[str, Any],
    ) -> Dict[str, Any]:
        """Create a new checkpoint."""
        request_data = {
            "job_id": None,
            "name": name,
            "attributes": attributes,
            "user_id": str(self.user_id),
            "secret_type": "credential",
            "secret_id": str(credential_id) if credential_id else None,
            "model_category": "USER",
            "form_category": "ORCA",
            "dist_json": dist_config,
            "vendor": vendor,
            "region": region,
            "storage_name": storage_name,
            "iteration": iteration,
            "files": files,
        }
        data = self.post(json=request_data)
        return data



# Copyright (c) 2022-present, FriendliAI Inc. All rights reserved.

"""PeriFlow Catalog Client."""

from __future__ import annotations

from string import Template
from typing import Any, Dict, List, Optional
from uuid import UUID

from periflow.client.base import (
    Client,
    GroupRequestMixin,
    ProjectRequestMixin,
    UserRequestMixin,
)
from periflow.enums import CatalogImportMethod


class CatalogClient(
    Client[UUID], GroupRequestMixin, ProjectRequestMixin, UserRequestMixin
):
    """Catalog client."""

    def __init__(self, **kwargs):
        """Initialize catalog client."""
        self.initialize_user()
        self.initialize_group()
        self.initialize_project()
        super().__init__(**kwargs)

    @property
    def url_path(self) -> Template:
        """Get an URL path."""
        return Template(self.url_provider.get_mr_uri("catalogs/"))

    def get_catalog(self, catalog_id: UUID) -> Dict[str, Any]:
        """Get a public checkpoint in catalog."""
        data = self.retrieve(pk=catalog_id)
        return data

    def list_catalogs(self, name: Optional[str], limit: int) -> List[Dict[str, Any]]:
        """List public checkpoints in catalog."""
        params = {
            "sort_by_popularity": "descending",
            "statuses": "Active",
        }
        if name is not None:
            params["name"] = name

        data = self.list(pagination=True, limit=limit, params=params)
        return data

    def try_out(
        self, catalog_id: UUID, name: str, method: CatalogImportMethod
    ) -> Dict[str, Any]:
        """Import a public checkpoint to a project.

        Args:
            catalog_id (UUID): ID of public checkpoint in catalog.
            name (str): The name of model to create in the project.
            method (CatalogImportMethod): Import method.

        Returns:
            Dict[str, Any]: Information of the created checkpoint.

        """
        request_data = {
            "name": name,
            "organization_id": str(self.group_id),
            "project_id": str(self.project_id),
            "user_id": str(self.user_id),
            "import_method": method.value,
        }
        data = self.post(path=f"{catalog_id}/try_out/", json=request_data)
        return data



# Copyright (c) 2022-present, FriendliAI Inc. All rights reserved.

"""PeriFlow CLI API Request Utilities."""

from __future__ import annotations

from requests.exceptions import HTTPError

from periflow.utils.url import discuss_url

DEFAULT_REQ_TIMEOUT = 30


def decode_http_err(exc: HTTPError) -> str:
    """Decode HTTP error."""
    try:
        if exc.response.status_code == 500:
            error_str = f"Internal Server Error: Please contact to system admin via {discuss_url}"
        elif exc.response.status_code == 404:
            error_str = (
                "Not Found: The requested resource is not found. Please check it again. "
                f"If you cannot find out why this error occurs, please visit {discuss_url}."
            )
        else:
            response = exc.response
            detail_json = response.json()
            if "detail" in detail_json:
                error_str = f"Error Code: {response.status_code}\nDetail: {detail_json['detail']}"
            elif "error_description" in detail_json:
                error_str = (
                    f"Error Code: {response.status_code}\n"
                    f"Detail: {detail_json['error_description']}"
                )
            else:
                error_str = f"Error Code: {response.status_code}"
    except ValueError:
        error_str = exc.response.content.decode()

    return error_str



# Copyright (c) 2022-present, FriendliAI Inc. All rights reserved.

"""PeriFlow Checkpoint Converter."""

from __future__ import annotations

from abc import ABC
from typing import Callable, Dict, List, Optional, Union

import h5py  # type: ignore[import]
import numpy as np
import torch
from tqdm.auto import tqdm
from transformers import GenerationConfig, PretrainedConfig  # type: ignore[import]

from periflow.enums import CheckpointDataType
from periflow.errors import NotSupportedCheckpointError
from periflow.logging import logger
from periflow.modules.converter.interface import (
    DecoderTFBlockConversionInterface,
    EncoderTFBlockConversionInterface,
    ModelConversionInterface,
    NonTFBlockConversionInterface,
)
from periflow.modules.converter.utils import (
    convert_tensor_to_np_array,
    get_tensor_from_state_dict,
)

SUPPORTED_GELU_FAMILY = [
    "gelu",
    "gelu_fast",
    "gelu_new",
    "gelu_python",
    "gelu_pytorch_tanh",
    "gelu_accurate",
]
SUPPORTED_HEAD_SIZE = [64, 80, 96, 128, 256]

ENCODER_PREFIX = "encoder"
DECODER_PREFIX = "decoder"


class AbstractConverter(ABC):
    """Abstract class for converting Hugging Face checkpoint to Periflow checkpoint.

    Attributes:
        config (PreTrainedConfig): Hugging Face model configuration.
        generation_config (Optional[GenerationConfig]): Hugginface generation config.
            When set to None, `config` is used for configuring generation.
        data_type (CheckpointDataType): Data type for the Periflow checkpoint.
        quantize (bool): Whether to quantize the Periflow checkpoint.

    """

    def __init__(
        self,
        config: PretrainedConfig,
        generation_config: Optional[GenerationConfig],
        data_type: CheckpointDataType,
    ) -> None:
        """Initialize converter."""
        self.config = config
        self.generation_config = generation_config
        self.data_type = data_type

    def get_eos_token_id(self) -> Optional[int]:
        """Get ID of EOS token."""
        generation_eos_token_id = None
        if self.generation_config is not None:
            generation_eos_token_id = self.generation_config.eos_token_id

        config_eos_token_id = self.config.eos_token_id

        if generation_eos_token_id is None:
            eos_token_id = config_eos_token_id
        else:
            if generation_eos_token_id != config_eos_token_id:
                logger.warn(
                    "'eos_token' is different in generation_config (%s) and config (%s). "
                    "Please fill the correct value.",
                    generation_eos_token_id,
                    config_eos_token_id,
                )
                eos_token_id = None
            else:
                eos_token_id = config_eos_token_id

        if eos_token_id is None:
            logger.warn(
                "'eos_token' cannot be automatically configured. "
                "Please fill in the field by yourself."
            )

        return eos_token_id

    def token_embed_weight_convert(
        self,
        state_dict: Dict[str, torch.Tensor],
        layer: str,
        per_layer_postfixes: List[str],
    ) -> np.ndarray:
        """Convert embedding layer's weight to PeriFlow format.

        Args:
            state_dict (Dict[str, torch.Tensor]): The state_dict of the original checkpoint.
            layer (str): The layer name of the original checkpoint.
            per_layer_postfixes (List[str]): The list of postfixes of the layer.

        Returns:
            The numpy array of the converted embedding weight.

        """
        assert len(per_layer_postfixes) == 1
        param = get_tensor_from_state_dict(state_dict, layer + per_layer_postfixes[0])
        return convert_tensor_to_np_array(param=param, data_type=self.data_type)

    def pos_embed_weight_convert(
        self,
        state_dict: Dict[str, torch.Tensor],
        layer: str,
        per_layer_postfixes: List[str],
    ) -> np.ndarray:
        """Convert position embedding layer's weight to PeriFlow format.

        Args:
            state_dict (Dict[str, torch.Tensor]): The state_dict of the original checkpoint.
            layer (str): The layer name of the original checkpoint.
            per_layer_postfixes (List[str]): The list of postfixes of the layer.

        Returns:
            The numpy array of the converted position embedding weight.
        """
        assert len(per_layer_postfixes) == 1
        param = get_tensor_from_state_dict(state_dict, layer + per_layer_postfixes[0])
        return convert_tensor_to_np_array(param=param, data_type=self.data_type)

    def head_weight_convert(
        self,
        state_dict: Dict[str, torch.Tensor],
        layer: str,
        per_layer_postfixes: List[str],
    ) -> np.ndarray:
        """Convert head layer's weight to PeriFlow format.

        Args:
            state_dict (Dict[str, torch.Tensor]): The state_dict of the original checkpoint.
            layer (str): The layer name of the original checkpoint.
            per_layer_postfixes (List[str]): The list of postfixes of the layer.

        Returns:
            The numpy array of the converted head weight.

        """
        assert len(per_layer_postfixes) == 1
        param = get_tensor_from_state_dict(state_dict, layer + per_layer_postfixes[0])
        return convert_tensor_to_np_array(param=param, data_type=self.data_type)

    def linear_weight_convert(
        self,
        state_dict: Dict[str, torch.Tensor],
        layer: str,
        per_layer_postfixes: List[str],
    ) -> np.ndarray:
        """Convert linear layer's weight to PeriFlow format.

        Args:
            state_dict (Dict[str, torch.Tensor]): The state_dict of the original checkpoint.
            layer (str): The layer name of the original checkpoint.
            per_layer_postfixes (List[str]): The list of postfixes of the layer.

        Returns:
            The numpy array of the converted linear weight.

        """
        assert len(per_layer_postfixes) == 1
        param = get_tensor_from_state_dict(state_dict, layer + per_layer_postfixes[0])
        param = param.transpose(0, 1)
        return convert_tensor_to_np_array(param=param, data_type=self.data_type)

    def linear_bias_convert(
        self,
        state_dict: Dict[str, torch.Tensor],
        layer: str,
        per_layer_postfixes: List[str],
    ) -> np.ndarray:
        """Convert linear layer's bias to PeriFlow format.

        Args:
            state_dict (Dict[str, torch.Tensor]): The state_dict of the original checkpoint.
            layer (str): The layer name of the original checkpoint.
            per_layer_postfixes (List[str]): The list of postfixes of the layer.

        Returns:
            The numpy array of the converted linear bias.

        """
        assert len(per_layer_postfixes) == 1
        param = get_tensor_from_state_dict(state_dict, layer + per_layer_postfixes[0])
        return convert_tensor_to_np_array(param=param, data_type=self.data_type)

    def ln_weight_convert(
        self,
        state_dict: Dict[str, torch.Tensor],
        layer: str,
        per_layer_postfixes: List[str],
    ) -> np.ndarray:
        """Convert layer norm layer's weight to PeriFlow format.

        Args:
            state_dict (Dict[str, torch.Tensor]): The state_dict of the original checkpoint.
            layer (str): The layer name of the original checkpoint.
            per_layer_postfixes (List[str]): The list of postfixes of the layer.

        Returns:
            The numpy array of the converted layer norm weight.

        """
        assert len(per_layer_postfixes) == 1
        param = get_tensor_from_state_dict(state_dict, layer + per_layer_postfixes[0])
        return convert_tensor_to_np_array(param=param, data_type=self.data_type)

    def ln_bias_convert(
        self,
        state_dict: Dict[str, torch.Tensor],
        layer: str,
        per_layer_postfixes: List[str],
    ) -> np.ndarray:
        """Convert layer norm layer's bias to PeriFlow format.

        Args:
            state_dict (Dict[str, torch.Tensor]): The state_dict of the original checkpoint.
            layer (str): The layer name of the original checkpoint.
            per_layer_postfixes (List[str]): The list of postfixes of the layer.

        Returns:
            The numpy array of the converted layer norm bias.

        """
        assert len(per_layer_postfixes) == 1
        param = get_tensor_from_state_dict(state_dict, layer + per_layer_postfixes[0])
        return convert_tensor_to_np_array(param=param, data_type=self.data_type)

    def qkv_weight_convert(
        self,
        state_dict: Dict[str, torch.Tensor],
        layer: str,
        per_layer_postfixes: List[str],
    ) -> np.ndarray:
        """Convert qkv layer's weight to PeriFlow format.

        In the original checkpoint, the qkv weight is stored as a single tensor or
        separated by three tensors. In the PeriFlow checkpoint, it is stored as a single tensor.

        Args:
            state_dict (Dict[str, torch.Tensor]): The state_dict of the original checkpoint.
            layer (str): The layer name of the original checkpoint.
            per_layer_postfixes (List[str]): The list of postfixes of the layer.

        Returns:
            The numpy array of the converted qkv weight.

        """
        params = [
            get_tensor_from_state_dict(state_dict, layer + postfix)
            for postfix in per_layer_postfixes
        ]
        param = torch.cat(params, dim=0)
        param = param.transpose(0, 1)
        return convert_tensor_to_np_array(param=param, data_type=self.data_type)

    def qkv_bias_convert(
        self,
        state_dict: Dict[str, torch.Tensor],
        layer: str,
        per_layer_postfixes: List[str],
    ) -> np.ndarray:
        """Convert qkv layer's bias to PeriFlow format.

        In the original checkpoint, the qkv weight is stored as a single tensor or
        separated by three tensors. In the PeriFlow checkpoint, it is stored as a single tensor.

        Args:
            state_dict (Dict[str, torch.Tensor]): The state_dict of the original checkpoint.
            layer (str): The layer name of the original checkpoint.
            per_layer_postfixes (List[str]): The list of postfixes of the layer.

        Returns:
           The numpy array of the converted qkv bias.

        """
        params = [
            get_tensor_from_state_dict(state_dict, layer + postfix)
            for postfix in per_layer_postfixes
        ]
        param = torch.cat(params, dim=0)
        return convert_tensor_to_np_array(param=param, data_type=self.data_type)


class DecoderOnlyConverter(
    AbstractConverter,
    ModelConversionInterface,
    NonTFBlockConversionInterface,
    DecoderTFBlockConversionInterface,
):
    """Converter for Decoder-Only models."""

    def check_config(self) -> None:
        """Check if a convertible form of the checkpoint from the decoder-only model config."""
        if self.decoder_head_size not in SUPPORTED_HEAD_SIZE:
            raise NotSupportedCheckpointError(
                invalid_option=f"decoder_head_size={self.decoder_head_size}",
                valid_options=SUPPORTED_HEAD_SIZE,
            )

    def get_convert_dict(
        self,
    ) -> Dict[str, Dict[str, Callable[[Dict[str, torch.Tensor], str], np.ndarray]]]:
        """Get convert dict for Decoder-Only model."""
        return {
            "non-transformer": self.non_transformer_convert_dict,
            "decoder": self.decoder_convert_dict,
        }

    def convert(
        self,
        model: torch.nn.Module,
        output_path: str,
        convert_dict: Dict[
            str, Dict[str, Callable[[Dict[str, torch.Tensor], str], np.ndarray]]
        ],
    ) -> None:
        """Convert Decoder-Only model's all layer to PeriFlow format."""
        state_dict = model.state_dict()
        total_layers = len(convert_dict["decoder"]) * self.decoder_layer_num + len(
            convert_dict["non-transformer"]
        )
        with h5py.File(output_path, "w") as out_f, tqdm(
            total=total_layers, desc="Converting", unit="tensor"
        ) as pbar:
            self.convert_decoder_layers(
                state_dict=state_dict,
                convert_dict=convert_dict["decoder"],
                out_f=out_f,
                pbar=pbar,
            )
            self.convert_non_transformer_layers(
                state_dict=state_dict,
                convert_dict=convert_dict["non-transformer"],
                out_f=out_f,
                pbar=pbar,
            )


class EncoderDecoderConverter(
    AbstractConverter,
    ModelConversionInterface,
    NonTFBlockConversionInterface,
    EncoderTFBlockConversionInterface,
    DecoderTFBlockConversionInterface,
):
    """Converter for Encoder-Decoder models."""

    def check_config(self) -> None:
        """Check if a convertible form of the checkpoint from the encoder-decoder model config."""
        if self.decoder_head_size not in SUPPORTED_HEAD_SIZE:
            raise NotSupportedCheckpointError(
                invalid_option=f"decoder_head_size={self.decoder_head_size}",
                valid_options=SUPPORTED_HEAD_SIZE,
            )

    def get_convert_dict(
        self,
    ) -> Dict[str, Dict[str, Callable[[Dict[str, torch.Tensor], str], np.ndarray]]]:
        """Get convert dict for Encoder-Decoder model."""
        return {
            "non-transformer": self.non_transformer_convert_dict,
            "encoder": self.encoder_convert_dict,
            "decoder": self.decoder_convert_dict,
        }

    def convert(
        self,
        model: torch.nn.Module,
        output_path: str,
        convert_dict: Dict[
            str, Dict[str, Callable[[Dict[str, torch.Tensor], str], np.ndarray]]
        ],
    ) -> None:
        """Convert Encoder-Decoder model's all layer to PeriFlow format."""
        state_dict = model.state_dict()
        total_layers = (
            len(self.encoder_convert_dict) * self.encoder_layer_num
            + len(self.decoder_convert_dict) * self.decoder_layer_num
            + len(self.non_transformer_convert_dict)
        )
        with h5py.File(output_path, "w") as out_f, tqdm(
            total=total_layers, desc="Converting", unit="tensor"
        ) as pbar:
            self.convert_decoder_layers(
                state_dict=state_dict,
                convert_dict=convert_dict["decoder"],
                out_f=out_f.create_group[DECODER_PREFIX],
                pbar=pbar,
            )
            self.convert_encoder_layers(
                state_dict=state_dict,
                convert_dict=convert_dict["encoder"],
                out_f=out_f.create_group[ENCODER_PREFIX],
                pbar=pbar,
            )
            self.convert_non_transformer_layers(
                state_dict=state_dict,
                convert_dict=convert_dict["non-transformer"],
                out_f=out_f,
                pbar=pbar,
            )

    def get_decoder_start_token_id(self) -> Optional[int]:
        """Get ID of decoder start token."""
        generation_decoder_start_token_id = None
        if self.generation_config is not None:
            generation_decoder_start_token_id = (
                self.generation_config.decoder_start_token_id
            )

        config_decoder_start_token_id = self.config.decoder_start_token_id

        if generation_decoder_start_token_id is None:
            decoder_start_token_id = config_decoder_start_token_id
        else:
            if generation_decoder_start_token_id != config_decoder_start_token_id:
                logger.warn(
                    "'decoder_start_token_id' is different in generation_config "
                    "(%s) and config (%s). Please fill the correct value.",
                    generation_decoder_start_token_id,
                    config_decoder_start_token_id,
                )
                decoder_start_token_id = None
            else:
                decoder_start_token_id = config_decoder_start_token_id

        if decoder_start_token_id is None:
            logger.warn(
                "'decoder_start_token' cannot be automatically configured. "
                "Please fill in the field by yourself."
            )

        return decoder_start_token_id


OneOfConverter = Union[EncoderDecoderConverter, DecoderOnlyConverter]



# Copyright (c) 2022-present, FriendliAI Inc. All rights reserved.

"""PeriFlow CodeGen Checkpoint Converter."""

from __future__ import annotations

from functools import partial
from typing import Any, Callable, Dict, List, cast

import numpy as np
import torch
from transformers import CodeGenConfig  # type: ignore[import]

from periflow.errors import CheckpointConversionError, NotSupportedCheckpointError
from periflow.logging import logger
from periflow.modules.converter.base import (
    DECODER_PREFIX,
    SUPPORTED_GELU_FAMILY,
    DecoderOnlyConverter,
)
from periflow.modules.converter.utils import (
    convert_tensor_to_np_array,
    get_tensor_from_state_dict,
    nontype_partial,
)


class CodegenForCausalLMConverter(DecoderOnlyConverter):
    """CodegenForCausalLM Architectures Converter Class."""

    def check_config(self) -> None:
        """Check if CodeGen architectures' config can be converted to PeriFlow format."""
        super().check_config()
        try:
            if (
                cast(CodeGenConfig, self.config).activation_function
                not in SUPPORTED_GELU_FAMILY
            ):
                raise NotSupportedCheckpointError(
                    invalid_option="'activation_function="
                    f"{cast(CodeGenConfig, self.config).activation_function}'",
                    valid_options=SUPPORTED_GELU_FAMILY,
                )
            if cast(CodeGenConfig, self.config).tie_word_embeddings:
                raise NotSupportedCheckpointError(
                    invalid_option="'tie_word_embeddings=True'",
                    valid_options=[False],
                )
            if cast(CodeGenConfig, self.config).layer_norm_epsilon != 1e-5:
                raise NotSupportedCheckpointError(
                    invalid_option="'layer_norm_epsilon="
                    f"{cast(CodeGenConfig, self.config).layer_norm_epsilon}'",
                    valid_options=[1e-5],
                )
        except AttributeError as exc:
            raise CheckpointConversionError(str(exc)) from exc

    def qkv_weight_convert(
        self,
        state_dict: Dict[str, torch.Tensor],
        layer: str,
        per_layer_postfixes: List[str],
    ) -> np.ndarray:
        """qkv_weight_convert for CodeGen's attention layer."""
        original_qkv_weight = get_tensor_from_state_dict(
            state_dict, layer + per_layer_postfixes[0]
        )
        reshaped_qkv_weight = original_qkv_weight.reshape(
            (4, original_qkv_weight.size(0) // 4, original_qkv_weight.size(1))
        )
        q_weight, v_weight, k_weight = torch.split(
            reshaped_qkv_weight, reshaped_qkv_weight.size(1) // 3, dim=1
        )
        q_weight = q_weight.reshape((-1, q_weight.size(2)))
        k_weight = k_weight.reshape((-1, k_weight.size(2)))
        v_weight = v_weight.reshape((-1, v_weight.size(2)))

        qkv_weight = torch.cat((q_weight, k_weight, v_weight), dim=0)
        qkv_weight = qkv_weight.transpose(0, 1)

        return convert_tensor_to_np_array(qkv_weight, self.data_type)

    def get_attributes(self) -> Dict[str, Any]:
        """Get checkpoint attributes."""
        config = cast(CodeGenConfig, self.config)

        logger.info(
            "The generated attributes set 'max_length' to %d, but you can change the "
            "'max_length' according to your needs. The CodeGen model does not rely on "
            "absolute position embeddings, allowing you to choose any suitable value.",
            config.n_positions,
        )

        eos_token_id = self.get_eos_token_id()
        attr = {
            "model_type": self.model_type,
            "dtype": self.data_type.value,
            "head_size": self.decoder_head_size,
            "rotary_dim": self.rotary_dim,
            "num_heads": self.decoder_num_attention_heads,
            "num_layers": self.decoder_layer_num,
            "max_length": config.n_positions,
            "vocab_size": config.vocab_size,
            "eos_token": eos_token_id if eos_token_id is not None else "FILL ME",
        }
        return attr

    @property
    def model_type(self) -> str:
        """Model type."""
        return "gpt-j"

    @property
    def non_transformer_convert_dict(
        self,
    ) -> Dict[str, Callable[[Dict[str, torch.Tensor], str], np.ndarray]]:
        """The convert_dict for non-transformer blocks in CodeGen."""
        return {
            "wte/weight:0": nontype_partial(
                self.token_embed_weight_convert,
                per_layer_postfixes=["transformer.wte.weight"],
            ),
            DECODER_PREFIX
            + "/ln_f/gamma:0": nontype_partial(
                self.ln_weight_convert, per_layer_postfixes=["transformer.ln_f.weight"]
            ),
            DECODER_PREFIX
            + "/ln_f/beta:0": nontype_partial(
                self.ln_bias_convert, per_layer_postfixes=["transformer.ln_f.bias"]
            ),
            "head_fc/weight:0": nontype_partial(
                self.head_weight_convert, per_layer_postfixes=["lm_head.weight"]
            ),
            "head_fc/bias:0": nontype_partial(
                self.linear_bias_convert, per_layer_postfixes=["lm_head.bias"]
            ),
        }

    @property
    def decoder_convert_dict(
        self,
    ) -> Dict[str, Callable[[Dict[str, torch.Tensor], str], np.ndarray]]:
        """The convert_dict for transformer blocks in CodeGen."""
        return {
            "ln_1/gamma:0": nontype_partial(
                self.ln_weight_convert,
                per_layer_postfixes=[".ln_1.weight"],
            ),
            "ln_1/beta:0": nontype_partial(
                self.ln_bias_convert,
                per_layer_postfixes=[".ln_1.bias"],
            ),
            "mlp/c_fc/bias:0": nontype_partial(
                self.linear_bias_convert,
                per_layer_postfixes=[".mlp.fc_in.bias"],
            ),
            "mlp/c_proj/bias:0": nontype_partial(
                self.linear_bias_convert,
                per_layer_postfixes=[".mlp.fc_out.bias"],
            ),
            "attn/c_attn/weight:0": nontype_partial(
                self.qkv_weight_convert,
                per_layer_postfixes=[".attn.qkv_proj.weight"],
            ),
            "attn/c_proj/weight:0": nontype_partial(
                self.linear_weight_convert,
                per_layer_postfixes=[".attn.out_proj.weight"],
            ),
            "mlp/c_fc/weight:0": nontype_partial(
                self.linear_weight_convert,
                per_layer_postfixes=[".mlp.fc_in.weight"],
            ),
            "mlp/c_proj/weight:0": nontype_partial(
                self.linear_weight_convert,
                per_layer_postfixes=[".mlp.fc_out.weight"],
            ),
        }

    @property
    def decoder_layer_prefix(self) -> str:
        """The layer name prefix used before CodeGen's transformer block number."""
        return "transformer.h."

    @property
    def decoder_layer_num(self) -> int:
        """The number of decoder layers in CodeGen."""
        return cast(CodeGenConfig, self.config).num_hidden_layers

    @property
    def decoder_hidden_size(self) -> int:
        """The hidden size in CodeGen."""
        return cast(CodeGenConfig, self.config).hidden_size

    @property
    def decoder_num_attention_heads(self) -> int:
        """The number of attention heads in CodeGen."""
        return cast(CodeGenConfig, self.config).num_attention_heads

    @property
    def decoder_num_kv_attention_heads(self) -> int:
        """The number of key-value attention heads in the codegen."""
        return self.decoder_num_attention_heads

    @property
    def decoder_head_size(self) -> int:
        """The head siez of CodeGen."""
        return self.decoder_hidden_size // self.decoder_num_attention_heads

    @property
    def rotary_dim(self) -> int:
        """The rotary dim in CodeGen."""
        return cast(CodeGenConfig, self.config).rotary_dim



# Copyright (c) 2022-present, FriendliAI Inc. All rights reserved.

"""PeriFlow Project CLI."""

from __future__ import annotations

from typing import Optional, Tuple
from uuid import UUID

import typer

from periflow.cli.group import get_current_org, get_org_user_id_by_email
from periflow.client.group import GroupProjectClient
from periflow.client.project import ProjectClient, find_project_id
from periflow.client.user import UserClient, UserGroupProjectClient
from periflow.context import (
    get_current_project_id,
    project_context_path,
    set_current_project_id,
)
from periflow.enums import GroupRole, ProjectRole
from periflow.formatter import PanelFormatter, TableFormatter
from periflow.utils.decorator import check_api
from periflow.utils.format import secho_error_and_exit

app = typer.Typer(
    no_args_is_help=True,
    context_settings={"help_option_names": ["-h", "--help"]},
    add_completion=False,
)
project_table_formatter = TableFormatter(
    name="Project", fields=["name", "id"], headers=["Name", "ID"]
)
project_panel_formatter = PanelFormatter(
    name="Project Detail",
    fields=["pf_group_id", "id", "name"],
    headers=["Organization ID", "Project ID", "Name"],
)
member_table_formatter = TableFormatter(
    name="Members",
    fields=["id", "name", "email", "access_level"],
    headers=["ID", "Name", "Email", "Role"],
)


# pylint: disable=redefined-builtin
@app.command()
@check_api
def list(
    tail: Optional[int] = typer.Option(
        None, "--tail", help="The number of project list to view at the tail"
    ),
    head: Optional[int] = typer.Option(
        None, "--tail", help="The number of project list to view at the head"
    ),
    show_group_project: bool = typer.Option(
        False, "--group", "-g", help="Show all projects in the current group"
    ),
):
    """List all accessible projects."""
    if show_group_project:
        client = GroupProjectClient()
    else:
        client = UserGroupProjectClient()  # type: ignore

    projects = client.list_projects()
    current_project_id = get_current_project_id()

    for project in projects:
        if current_project_id is not None and project["id"] == str(current_project_id):
            project["name"] = f"[bold green]* {project['name']}"
            project["id"] = f"[bold green]{project['id']}"
        else:
            project["name"] = f"  {project['name']}"

    if tail is not None or head is not None:
        target_project_list = []
        if tail is not None:
            target_project_list.extend(projects[:tail])
        if head is not None:
            target_project_list.extend(
                projects[-head:]  # pylint: disable=invalid-unary-operand-type
            )
    else:
        target_project_list = projects

    project_table_formatter.render(target_project_list)


@app.command()
@check_api
def create(name: str = typer.Argument(..., help="Name of project to create")):
    """Create a new project.

    :::info
    Every resource within a project, such as credentials, checkpoints, and deployments,
    is shared with project members.
    :::

    """
    client = GroupProjectClient()
    project_detail = client.create_project(name)
    project_panel_formatter.render(project_detail)


@app.command()
@check_api
def current():
    """Get the current working project."""
    client = ProjectClient()
    project_id = get_current_project_id()
    if project_id is None:
        secho_error_and_exit("working project is not set")
    project = client.get_project(project_id)
    project_panel_formatter.render(project)


@app.command()
@check_api
def switch(
    name: str = typer.Argument(
        ...,
        help="Name of project to switch",
    )
):
    """Switch working project context.

    There can exist multiple projects under the organization, and each organization
    member can be included in multiple projects. You first need to select the working
    project context where you want to use the services.

    :::info
    Your working project context is written at `./periflow/project` file in your home directory.
    :::

    """
    user_group_project_client = UserGroupProjectClient()
    project_client = ProjectClient()

    project_id = find_project_id(user_group_project_client.list_projects(), name)
    if project_client.check_project_membership(pf_project_id=project_id):
        set_current_project_id(project_id)
        typer.secho(f"Project switched to {name}.", fg=typer.colors.BLUE)
    else:
        secho_error_and_exit(
            f"You don't have permission to project ({name}). Please contact to the project admin."
        )


@app.command()
@check_api
def delete(
    name: str = typer.Argument(
        ...,
        help="Name of project to delete",
    )
):
    """Delete the project."""
    project_client = ProjectClient()
    user_group_project_client = UserGroupProjectClient()
    project_id = find_project_id(user_group_project_client.list_projects(), name)
    project_client.delete_project(pf_project_id=project_id)
    if project_id == get_current_project_id():
        project_context_path.unlink()
    typer.secho(f"Project {name} deleted.", fg=typer.colors.BLUE)


@app.command("add-user")
@check_api
def add_user(
    email: str = typer.Argument(
        ...,
        help="Email of the user to add to the current working project",
    ),
    role: ProjectRole = typer.Argument(
        ...,
        help="Project role to assign",
    ),
):
    """Add user to project.

    If you are the **Admin** of a project, you can add members in the same organization
    to the project and assign their roles in the project.

    :::info
    If you have **Owner** role in the organization, you have full-access to all projects
    inside the organization.
    :::

    """
    user_client = UserClient()

    org_id, project_id = _check_project_and_get_id()
    user_id = get_org_user_id_by_email(org_id, email)

    user_client.add_to_project(user_id, project_id, role)
    typer.secho(
        f"User '{email}' is successfully added to project", fg=typer.colors.BLUE
    )


@app.command("delete-user")
@check_api
def delete_user(
    email: str = typer.Argument(
        ...,
        help="Email of the user to delete from the current working project",
    ),
    force: bool = typer.Option(
        False,
        "--force",
        "-f",
        help="Forcefully delete without confirmation prompt",
    ),
):
    """Delete user from the project.

    :::info
    Only organization **Owner** or project **Admin** can evict members from the project.
    :::

    """
    user_client = UserClient()

    org_id, project_id = _check_project_and_get_id()
    user_id = get_org_user_id_by_email(org_id, email)

    if not force:
        do_delete = typer.confirm(
            f"Are you sure to remove user '{email}' from the project?"
        )
        if not do_delete:
            raise typer.Abort()

    user_client.delete_from_project(user_id, project_id)
    typer.secho(
        f"User '{email}' is successfully deleted from project", fg=typer.colors.BLUE
    )


@app.command("set-role")
@check_api
def set_role(
    email: str = typer.Argument(
        ...,
        help="Email of the user to assign a project role",
    ),
    role: ProjectRole = typer.Argument(
        ...,
        help="Project role",
    ),
):
    """Set project role for the user.

    Refer to the following table for the project-level roles and privileges.

    | | Admin | Maintainer | Developer | Guest |
    |-|-|-|-|-|
    | Add organization members to the project | ✓ | ✗ | ✗ | ✗ |
    | Assign project roles | ✓ | ✗ | ✗ | ✗ |
    | Delete project | ✓ | ✗ | ✗ | ✗ |
    | Delete & edit checkpoints | ✓ | ✗ | ✗ | ✗ |
    | Delete & edit credentials, deployments | ✓ | ✓ | ✗ | ✗ |
    | Creaete credentials | ✓ | ✓ | ✗ | ✗ |
    | Create checkpoints, deployments | ✓ | ✓ | ✓ | ✗ |
    | Read checkpoints, credentials, deployments | ✓ | ✓ | ✓ | ✓ |
    | Send inference requests to deployments | ✓ | ✓ | ✓ | ✓ |

    :::info
    Only the **Admin** users can assign roles to project members:
    :::

    :::info
    If you have the **Owner** role in the organization, you have full-access to all
    projects inside the organization.
    :::

    """
    user_client = UserClient()

    org_id, project_id = _check_project_and_get_id()
    user_id = get_org_user_id_by_email(org_id, email)

    user_client.set_project_privilege(user_id, project_id, role)
    typer.secho(
        f"Project role for user '{email}' successfully updated to {role.value}!"
    )


@app.command()
@check_api
def members():
    """List project members."""
    project_client = ProjectClient()

    project_id = get_current_project_id()
    if project_id is None:
        secho_error_and_exit("Failed to identify project... Please set project again.")

    members = project_client.list_users(project_id)
    member_table_formatter.render(members)


def _check_project_and_get_id() -> Tuple[UUID, UUID]:
    """Get org_id and project_id if valid."""
    user_client = UserClient()

    org = get_current_org()
    project_id = get_current_project_id()
    if project_id is None:
        secho_error_and_exit("Failed to identify project... Please set project again.")

    if org["privilege_level"] == GroupRole.OWNER:
        return UUID(org["id"]), project_id

    requester = user_client.get_project_membership(project_id)
    if requester["access_level"] != ProjectRole.ADMIN:
        secho_error_and_exit("Only the admin of the project can add-user/set-role")

    return org["id"], project_id
